#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Script(DbDict):
    '''This is a "ninja" class.  There are no Script resources in Bacula, but there should be.

    Scripts handle the RunScript items in Jobs.
    '''
    SETUP_KEYS = [(RUNSWHEN, 'Never', 'Legal values are "Never", "Before", "After", "AfterVSS", and "Always"'), COMMAND, CONSOLE]
    BOOL_KEYS = [RUNSONSUCCESS, RUNSONCLIENT, FAILJOBONERROR, RUNSONFAILURE,]
    table = SCRIPTS
    prefix = '    '
    # {{{ __str__(): 

    def __str__(self):
        '''Standardized string representation of a Script, suitable for including in a Job configuration.'''
        self.output = ['  RunScript {  # Script ID: %d\n    # Name: %s' % (self[ID], self[NAME]),'  }']
        for key in self.SETUP_KEYS: self._simple_phrase(key)
        for key in self.BOOL_KEYS: self._yesno_phrase(key)
        return '\n'.join(self.output)

    # }}}
    # {{{ search():

    def search(self, key=None):
        '''Since Scripts are ninjas, they have no Name=FOO in the configuration
        file.  We will generate names upon creation, and use them, but we have to
        support much more complicated searching to deal with the reality of Bacula
        configurations.'''
        debug_print('dbdict search')
        DbDict.search(self, key)
        if self[ID]: return self
        debug_print('continuing search')
        conditions = []
        values = []
        for key in self.keys():
            if not self[key] == None:
                conditions.append('`%s` = %%s' % key)
                values.append(self[key])
        sql = 'select * from %s where %s' % (self.table, ' and '.join(conditions))
        values = tuple(values)

        new_me = self.bc.do_sql(sql, values, asdict=True)
        if not new_me: self._create()
        else: self.update(new_me[0])
        return self

    # }}}
    # {{{ _create():

    def _create(self):
        '''This is where new Scripts get created and named.'''
        keys = []
        keysub=[]
        values = []
        for key in self.keys():
            if not self[key] == None:
                keys.append('`%s`' % key)
                keysub.append('%s')
                values.append(self[key])
        sql = 'insert into %s (%s) values (%s)' % (self.table, ','.join(keys), ','.join(keysub))
        self.bc.do_sql(sql, tuple(values), asdict=True)
        self.search()
        self[NAME] = 'Script: %d' % self[ID]
        self._save()
        return self

    # }}}

def main():
    s = Script()
    s.cli()

if __name__ == "__main__": main()



from . import *

# client for interacting with a storage daemon

class SDaemon(BSock):
    '''Client for communicating directly with a storage daemon.
    '''
# {{{ __init__(address, password, myname, port=BACULA_SD_PORT, debug=False, timeout=5):
    def __init__(self, address, password, myname, port=BACULA_SD_PORT, debug=False, timeout=5):
        BSock.__init__(self, address, password, 'Director ' + myname,
                                             port, debug, timeout)
        return
# }}}
# {{{ version(): request version info from the remote daemon (useless?)

    def version(self):
        '''Query the SD for its version.'''
        for line in self.status().split('\n'):
            if 'Version' in line: return line
        return 'Unable to determine version'

    # }}}
# {{{ sd_key(jobid=0, job=None, sdid=0, sdtime=0, authorization='dummy'): SD key setup 

    def sd_key(self, jobid=0, job=None, sdid=0, sdtime=0, authorization='dummy'):
        '''You need an SD key before you can actually do anything with an FD/SD'''
        if not job: job='-Console-.' + self._time()
        self.send('JobId=%(jobid)d Job=%(job)s SDid=%(sdid)d SDtime=%(sdtime)d Authorization=%(authorization)s' % locals())
        return self.recv()

# }}}



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Console(DbDict):
    '''This is for configuring bconsole access.  Unfortunately, there's no good
    way to extract this for constructing a bconsole.conf.  You can set up
    conventions that a client hostname should match the name of te Console,
    but that isn't going to be very satisfactory.

    '''
    SETUP_KEYS = [CATALOGACL, CLIENTACL, COMMANDACL, FILESETACL, JOBACL, PASSWORD,
                  POOLACL, SCHEDULEACL, STORAGEACL, WHEREACL]
    table = CONSOLES
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import quotedString, restOfLine, Keyword, nestedExpr, OneOrMore, Word, Literal, removeQuotes
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        for key in self.NULL_KEYS:
            if key == id: continue
            gr_line = gr_line | np((key,), action=self._parse_setter(key))

        gr_res = OneOrMore(gr_line)
        result = gr_res.parseString(string, parseAll=True)
        return 'Console: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        self.output = ['Console {\n  Name = "%(name)s"' % self, '}']
        
        for key in self.SETUP_KEYS:
            self._simple_phrase(key)
        return '\n'.join(self.output)

# }}}


# Implement the CLI for managing Consoles
def main():
    s = Console()
    s.cli()

if __name__ == "__main__": main()



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *
from re import compile, MULTILINE, IGNORECASE, DOTALL

class Fileset(DbDict):
    table = FILESETS
    BOOL_KEYS = [ENABLEVSS, IGNOREFILESETCHANGES,]
    def __init__(self, row={}, string=None):
        DbDict.__init__(self, row, string)
        self.entries = []
        return
    # {{{ _load_parts(): helper for loading self from the database

    def _load_parts(self):
        '''Load up the rules from the table.'''
        if not self[ID]: return
        sql = '''SELECT b.id AS id, b.name AS name, b.option, a.exclude
                 FROM fileset_link a, fileset_files b
                 WHERE a.file_id = b.id AND a.fileset_id = %s'''
        self.entries = list(self.bc.do_sql(sql, self[ID]))
        return self

    # }}}
    # {{{ _parse_add_entry(*args): adds various entries into this fileset

    def _parse_add_entry(self, *args):
        '''Supports the string parser.  Extracts fileset Include/Exclude phrases.
        This should probably be thrown back into the parser, but I'll leave
        that for a later effort.  For now, it works just fine.
        '''
        trim = args[2]
        if trim[0] == EXCLUDE: exclude=1
        else: exclude=0
        trim = trim[1]
        flag = args[2][0][0]
        data = args[2][0][1][0].strip()
        for row in trim:
            flag = row[0]
            if row[0] == OPTIONS:
                for opt in row[1][0].strip().split('\n'):
                    self._add_entry(opt.strip(), 1, exclude)
            else:
                self._add_entry(' '.join([x.strip() for x in row]), 0, exclude)
        return

    # }}}
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_name = Keyword('name', caseless=True) + gr_eq + gr_opt_quoted_string
        gr_name.setParseAction(lambda x, y=self: y._set_name(x[2]))
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))
        gr_phrase = Group(OneOrMore(gr_stripped_string | Word(alphanums)) + gr_eq + gr_opt_quoted_string)

        def np(words, fn = gr_opt_quoted_string, action=print):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p
        
        gr_ifsc = np(PList('Ignore File Set Changes'), gr_yn, action=self._parse_setter(IGNOREFILESETCHANGES))
        gr_evss = np(PList('Enable VSS'), gr_yn, action=self._parse_setter(ENABLEVSS))

        gr_i_option = Group(Keyword(OPTIONS, caseless=True) + nestedExpr('{','}', Regex('[^\}]+', re.MULTILINE)))
        gr_e_option = gr_i_option.copy()
        gr_i_file = gr_phrase.copy()
        gr_e_file = gr_phrase.copy()

        gr_inc = Keyword('include', caseless=True) + nestedExpr('{','}', OneOrMore(gr_i_option | gr_i_file))
        gr_inc.addParseAction(self._parse_add_entry)
        gr_exc = Keyword('exclude', caseless=True) + nestedExpr('{','}', OneOrMore(gr_e_option | gr_e_file))
        gr_exc.addParseAction(self._parse_add_entry)

        gr_res = OneOrMore(gr_name | gr_inc | gr_exc | gr_ifsc | gr_evss)
        result = gr_res.parseString(string, parseAll=True)
        return 'Fileset: ' + self[NAME]

    # }}}
    # {{{ _add_entry(entry, option, exclude): add another entry

    def _add_entry(self, entry, option, exclude):
        '''This holds the SQL that manages writing entries.  It probably needs a
        complete redesign.
        
        The logic here could bite you if it were legal to have an option
        and a file with the same content.

        '''
        new_entry = list(self.bc.value_ensure(FILESET_FILES, NAME, entry)[0])
        if not new_entry[2] == option:
            new_entry[2] = option
            list(self.bc.do_sql('UPDATE fileset_files SET `option` = %s WHERE id = %s', (new_entry[2], new_entry[0])))

        new_entry.append(exclude)
        row = self.bc.do_sql('SELECT * FROM fileset_link WHERE fileset_id = %s AND file_id = %s and exclude = %s',
                             (self[ID], new_entry[0], new_entry[3]))
        if not row:
            try: self.bc.do_sql('INSERT INTO fileset_link(fileset_id, file_id, exclude) VALUES (%s, %s, %s)',
                                (self[ID], new_entry[0], new_entry[3]))
            except: print('You may not have the same entry in both Include{} and Exclude{} clauses.')

        self._load_parts()
        return

    # }}}
    # {{{ _delete_entry(entry): delete an entry

    def _delete_entry(self, entry):
        '''Remove an entry.'''
        for row in self.entries:
            if not row[1] == entry: continue
            self.bc.do_sql('DELETE FROM fileset_link WHERE fileset_id = %s AND file_id = %s and exclude = %s',
                           (self[ID], row[0], row[3]))
            self.entries.remove(row)
            return
        print('I cannot delete entries that do not exist!')
        return

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''Stringify into a form suitable for dropping into a Director configuration.'''
        self.output = ['Fileset {','  Name = "%(name)s"' % self, '}']
        for key in self.BOOL_KEYS: self._yesno_phrase(key)

        for test,phrase in ([0,'Include'],[1,'Exclude']):
            subset =  [x for x in self.entries if x[3] == test]
            if subset:              # We have includes
                self.output.insert(-1, '  %s {' % phrase)
                options = [x for x in subset if x[2]]
                if options:
                    self.output.insert(-1,'    Options {')
                    [self.output.insert(-1,'      ' + x[1]) for x in options]
                    self.output.insert(-1,'    }')
                [self.output.insert(-1,'    %s' % x[1]) for x in subset if not x[2]]
                self.output.insert(-1,'  }')
        return '\n'.join(self.output)

# }}}
    # {{{ _cli_special_setup(): setup the weird phrases that go with filesets

    def _cli_special_setup(self):
        '''Add CLI options to add file/option inclusion/exclusion rules.'''
        group = optparse.OptionGroup(self.parser,
                                     "Stanzas",
                                     "Handle elements of the Include and Exclude sections")
        group.add_option('--add-file', action='append', default=[], help='Add a file to the Include section')
        group.add_option('--add-option', action='append', default=[], help='Add an option to the Include section')
        group.add_option('--add-exclusion-file', action='append', default=[], help='Add a file to the Exclude section')
        group.add_option('--add-exclusion-option', action='append', default=[], help='Add an option to the Exclude section')
        group.add_option('--remove', action='append', default=[], help='Removes exact text match for either an option or a file')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle the weird phrases that go with filesets

    def _cli_special_do_parse(self, args):
        '''Handle the args for adding/removing entries'''
        for stanza in args.add_file: self._add_entry(stanza, 0, 0)
        for stanza in args.add_option: self._add_entry(stanza, 1, 0)
        for stanza in args.add_exclusion_file: self._add_entry(stanza, 0, 1)
        for stanza in args.add_exclusion_option: self._add_entry(stanza, 1, 1)
        for stanza in args.remove: self._delete_entry(stanza)
        return

# }}}
    # {{{ _cli_special_print(): print out the weird phrases that go with filesets

    def _cli_special_print(self):
        '''Print out the Inclusion and Exclusion rules'''
        fmt = '%' + str(self._maxlen) + 's: %s'
        for test,phrase in ([0,'Include'],[1,'Exclude']):
            subset =  [x for x in self.entries if x[3] == test]
            if subset:              # We have a group!
                print(phrase)
                options = [x for x in subset if x[2]]
                if options:
                    print(fmt % ('Options', ','.join([x[1] for x in options])))
                [print(fmt % ('',x[1])) for x in subset if not x[2]]
                
    # }}}
    # {{{ _cli_special_clone(oid):

    def _cli_special_clone(self, oid):
        '''When cloning, ensure that all of the rules get copied into the cloned Fileset.'''
        select = 'SELECT %s,file_id, exclude FROM fileset_link WHERE fileset_id = %%s' % self[ID]
        insert = 'INSERT INTO fileset_link (fileset_id,file_id,exclude) VALUES (%s,%s,%s)'
        for row in self.bc.do_sql(select, oid): self.bc.do_sql(insert, row)
        self._load_parts()
        return

# }}}
        

def main():
    s = Fileset()
    s.cli()

if __name__ == "__main__": main()



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Pool(DbDict):
    BOOL_KEYS = [AUTOPRUNE, CATALOGFILES, RECYCLE, USEVOLUMEONCE,
                 RECYCLEOLDESTVOLUME, RECYCLECURRENTVOLUME, PURGEOLDESTVOLUME]
    SETUP_KEYS = [MAXIMUMVOLUMES, STORAGE, MAXIMUMVOLUMEJOBS, MAXIMUMVOLUMEFILES,
                  MAXIMUMVOLUMEBYTES, VOLUMEUSEDURATION, VOLUMERETENTION,
                  ACTIONONPURGE, SCRATCHPOOL, RECYCLEPOOL,
                  FILERETENTION, JOBRETENTION, CLEANINGPREFIX,
                  LABELFORMAT, (POOLTYPE, 'Backup', "Enum: ['Backup', 'Archive', 'Cloned', 'Migration', 'Copy', 'Save']")]
    table = POOLS
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        gr_line = gr_line | np(PList('pool type'), action=self._parse_setter(POOLTYPE))
        gr_line = gr_line | np(PList('maximum volumes'), action=self._parse_setter(MAXIMUMVOLUMES))
        gr_line = gr_line | np((STORAGE,), action=self._parse_setter(STORAGE))
        gr_line = gr_line | np(PList('use volume once'), gr_yn, action=self._parse_setter(USEVOLUMEONCE))
        gr_line = gr_line | np(PList('catalog files'), gr_yn, action=self._parse_setter(CATALOGFILES))
        gr_line = gr_line | np(PList('auto prune'), gr_yn, action=self._parse_setter(AUTOPRUNE))
        gr_line = gr_line | np((RECYCLE,), gr_yn, action=self._parse_setter(RECYCLE))
        gr_line = gr_line | np(PList('recycle oldest volume'), gr_yn, action=self._parse_setter(RECYCLEOLDESTVOLUME))
        gr_line = gr_line | np(PList('recycle current volume'), gr_yn, action=self._parse_setter(RECYCLECURRENTVOLUME))
        gr_line = gr_line | np(PList('purge oldest volume'), gr_yn, action=self._parse_setter(PURGEOLDESTVOLUME))
        gr_line = gr_line | np(PList('maximum volume jobs'), gr_number, action=self._parse_setter(MAXIMUMVOLUMEJOBS))
        gr_line = gr_line | np(PList('maximum volume files'), gr_number, action=self._parse_setter(MAXIMUMVOLUMEFILES))
        gr_line = gr_line | np(PList('maximum volume bytes'), action=self._parse_setter(MAXIMUMVOLUMEBYTES))
        gr_line = gr_line | np(PList('volume use duration'), action=self._parse_setter(VOLUMEUSEDURATION))
        gr_line = gr_line | np(PList('volume retention'), action=self._parse_setter(VOLUMERETENTION))
        gr_line = gr_line | np(PList('action on purge'), action=self._parse_setter(ACTIONONPURGE))
        gr_line = gr_line | np(PList('scratch pool'), action=self._parse_setter(SCRATCHPOOL))
        gr_line = gr_line | np(PList('recycle pool'), action=self._parse_setter(RECYCLEPOOL))
        gr_line = gr_line | np(PList('file retention'), action=self._parse_setter(FILERETENTION))
        gr_line = gr_line | np(PList('job retention'), action=self._parse_setter(JOBRETENTION))
        gr_line = gr_line | np(PList('cleaning prefix'), action=self._parse_setter(CLEANINGPREFIX))
        gr_line = gr_line | np(PList('label format'), action=self._parse_setter(LABELFORMAT))

        gr_res = OneOrMore(gr_line)
        result = gr_res.parseString(string, parseAll=True)
        return 'Pool: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Pool formatted for inclusion in a config file.'''
        self.output = ['Pool {\n  Name = "%(name)s"' % self,'}']
        for key in self.SETUP_KEYS: self._simple_phrase(key)
        for key in self.BOOL_KEYS: self._yesno_phrase(key)

        return '\n'.join(self.output)

# }}}

def main():
    s = Pool()
    s.cli()

if __name__ == "__main__": main()



'''Bacula configuration database stuff: common routines, credentials, etc.
Configuration is at the *end* of this file.'''

from __future__ import print_function
import re, shlex, sys, os

# Just a placeholder, it gets overridden later
DEBUG=False

# Bacula CONFIG DB bits that are less easily interned
BACULA_DIR_PORT = 9101
BACULA_FD_PORT = 9102
BACULA_SD_PORT = 9103

# Configuration files
BACULA_DIR_CONF = '/etc/bacula/bacula-dir.conf'
BACULA_FD_CONF = '/etc/bacula/bacula-fd.conf'
BACULA_SD_CONF = '/etc/bacula/bacula-sd.conf'
BCONSOLE_CONF = '/etc/bacula/bconsole.conf'

# MySQL credentials for the Bacula configuration database. They should be
# overridden in one of the configuration files (see below for the config
# files that are checked).
MYSQL_DB = 'OVERRIDE ME'
MYSQL_HOST = 'OVERRIDE ME'
MYSQL_USER = 'OVERRIDE ME'
MYSQL_PASS = 'OVERRIDE ME'

_INTERNED = [
    'Append', 'Available', 'Catalog', 'Cleaning', 'Error', 'Full', 'Purged', 'Recycle', 'Used',
    'actiononpurge', 'address', 'autoprune', 'autoprune', 'bacula_enabled', 'begin', 'catalog_id',
    'catalogacl', 'catalogfiles', 'catalogs', 'cleaningprefix', 'client', 'clientacl', 'clients',
    'commandacl', 'console', 'consoles', 'data', 'db', 'dbaddress', 'dbname', 'dbpassword',
    'dbport', 'dbsocket', 'dbuser', 'diraddresses', 'director', 'director_id', 'director_name',
    'directors', 'dirid', 'enabled', 'end', 'entries', 'exclude', 'excludes', 'failure',
    'fd_connect_timeout', 'fdaddress', 'fdaddresses', 'fdport', 'fdsourceaddress', 'file_retention',
    'filedaemon', 'fileretention', 'fileset', 'fileset_files', 'filesetacl', 'filesets',
    'heartbeatinterval', 'hostid', 'hostname', 'hostnames', 'hosts', 'id',
    'ignorefilesetchanges', 'includes', 'job_retention', 'jobacl', 'jobretention', 'labelformat',
    'lastupdated', 'maximumconcurrentjobs', 'maximumbandwidthperjob', 'maximumconcurrentjobs',
    'maximumconsoleconnections', 'maximumconsoleconnections', 'maximumnetworkbuffersize',
    'maximumvolumebytes', 'maximumvolumefiles', 'maximumvolumejobs', 'maximumvolumes', 'messages_id',
    'messages', 'monitor', 'name', 'no', 'notes', 'option', 'options', 'os', 'owners', 'password',
    'piddirectory', 'pkiencryption', 'pkikeypair', 'pkimasterkey', 'pkisignatures', 'pool',
    'poolacl', 'pools', 'pooltype', 'port', 'primary_dir', 'priority', 'purgeoldestvolume',
    'queryfile', 'recycle', 'recyclecurrentvolume', 'recycleoldestvolume', 'recyclepool', 'rows',
    'run', 'schedule', 'schedule_time', 'scheduleacl', 'schedules', 'scratchpool', 'sdport',
    'scripts_directory', 'sd_connect_timeout', 'sdconnecttimeout', 'service', 'services',
    'sourceaddress', 'statistics_retention', 'storage', 'storageacl', 'storagepassword',
    'storageserver', 'storageserveraddress', 'timespan', 'user', 'usevolumeonce', 'verid',
    'volumeretention', 'volumeuseduration', 'enablevss', 'whereacl', 'workingdirectory', 'yes',
    'device', 'mediatype', 'autochanger', 'allowcompression', 'type', 'level', 'differentialpool_id',
    'fileset_id', 'fullpool_id', 'incrementalpool_id', 'client_id', 'pool_id',
    'schedule_id', 'job_id', 'rescheduletimes', 'accurate', 'allowduplicatejobs',
    'allowmixedpriority', 'cancellowerlevelduplicates', 'cancelqueuedduplicates',
    'cancelrunningduplicates', 'prefermountedvolumes', 'prefixlinks', 'prunefiles', 'prunejobs',
    'prunevolumes', 'rerunfailedlevels', 'rescheduleonerror', 'spoolattributes', 'spooldata',
    'writepartafterjob', 'addprefix', 'addsuffix', 'base', 'bootstrap', 'differentialmaxwaittime',
    'idmaxwaittime', 'incrementalmaxruntime', 'maxfullinterval', 'maximumbandwidth',
    'maxrunschedtime', 'maxruntime', 'maxstartdelay', 'maxwaittime', 'regexwhere',
    'rescheduleinterval', 'spoolsize', 'stripprefix', 'verifyjob', 'where', 'writebootstrap',
    'replace', 'jobdef', 'jobs', 'job', 'storage_id', 'command', 'runsonsuccess', 'runsonfailure',
    'runsonclient', 'runswhen', 'failjobonerror', 'scripts', 'message', 'script_id', 'jobdefs',
    'clientconnectwait', 'sdaddresses', 'sdaddress',

    'archivedevice', 'devicetype', 'changerdevice', 'changercommand', 'alertcommand',
    'driveindex', 'maximumchangerwait', 'maximumrewindwait',
    'maximumopenwait', 'volumepollinterval', 'mountpoint', 'mountcommand', 'unmountcommand',
    'minimumblocksize', 'maximumblocksize', 'maximumvolumesize', 'maximumfilesize',
    'maximumnetworkbuffersize', 'maximumspoolsize', 'maximumjobspoolsize', 'spooldirectory',
    'maximumpartsize', 'autoselect', 'alwaysopen', 'closeonpoll', 'removablemedia',
    'randomaccess', 'blockchecksum', 'hardwareendofmedium', 'fastforwardspacefile', 'usemtiocget',
    'bsfateom', 'twoeof', 'backwardspacerecord', 'backwardspacefile', 'forwardspacerecord',
    'forwardspacefile', 'offlineonunmount', 'blockpositioning', 'labelmedia', 'automaticmount',
    'clientconnectwait','fd','sd', 'bconsole', 'dirport', 'generate'
    ]

for w in _INTERNED: locals()[w.upper()] = w

STATUS = [FULL, USED, APPEND, CLEANING, ERROR, PURGED, RECYCLE, AVAILABLE]
TRUE_VALUES = ['1', 'yes', 'y', 'on', 'true', 't']
FALSE_VALUES = ['0', 'no', 'n', 'off', 'false', 'f']

WORKING_DIR = {
    'Linux': "/var/lib/bacula",
    'OSX': "/var/db/bacula",
    'FreeBSD': "/var/db/bacula",
    'Windows': "/bacula/working",
    }
    
def debug_print(msg, *args):
    global DEBUG
    if DEBUG: print(msg % args, file=sys.stderr)
    sys.stderr.flush()
    return

def set_debug(value):
    global DEBUG
    DEBUG = value
    return

set_debug(os.environ.get('DEBUG', False))

# These rules will be used to determine what fileset(s) and schedule(s)
# will be applied to new jobs that are generated in an automated fashion.
# Currently only used by webconfig.  Only HOSTNAME and OS are currently
# valid values.

# The rules in here are examples only!

# [variable, regex, fileset, schedule]
_guessing_rules = [
    (HOSTNAME, re.compile('origin'), 'SetOnHostGZIP', 'CustomHost'),
    (HOSTNAME, re.compile('origin'), 'XferLogs', 'FtpHosts'),
    (HOSTNAME, re.compile(r'\.ocs\.'), 'SetOnHostGZIP', 'Weekly'),
    (OS, re.compile(r'Windows'), 'WinFullGZIP', 'Weekly'),
    ]
_default_rules = [('Daily', 'FullUnixGZIP')]

# Where we keep
BACULADATADIR = '/data/bacula'


# Now that we've set all of the default initial values, we'll load and
# parse config files, updating the environment from them.
CUSTOM_LIST = ['/etc/bacula/bacula.conf',
               '/usr/local/etc/bacula/bacula.conf',
               '/usr/local/etc/bacula.conf',
               os.path.join(os.environ['HOME'], '.bacula.conf')
               ]
for filename in CUSTOM_LIST:
    if not os.access(filename, os.R_OK): continue
    exec( open(filename).read(), locals(), locals())

from bacula_config import *
from util import *
from schedule import Schedule
from fileset import Fileset
from messages import Messages
from director import Director
from catalog import Catalog
from console import Console
from client import Client
from pool import Pool
from storage import Storage
from job import Job, JobDef
from scripts import Script
from device import Device

# bconsole/daemon bits
from fd import FDaemon
from sd import SDaemon
from bacula_director import BDirector


import util

_DISPATCHER = {
    FILESET: Fileset,
    SCHEDULE: Schedule,
    MESSAGES: Messages,
    MESSAGE: Messages,
    DIRECTOR: Director,
    CATALOG.lower(): Catalog,
    CONSOLE: Console,
    CLIENT: Client,
    FILEDAEMON: Client,
    POOL: Pool,
    STORAGE: Storage,
    JOB: Job,
    JOBDEF: JobDef,
    JOBDEFS: JobDef,
    DEVICE: Device,
    }



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Client(DbDict):
    SETUP_KEYS = [
        ADDRESS, CATALOG_ID, FILERETENTION, JOBRETENTION, 
        WORKINGDIRECTORY, SDCONNECTTIMEOUT, MAXIMUMNETWORKBUFFERSIZE,
        PIDDIRECTORY, HEARTBEATINTERVAL, FDADDRESSES, FDADDRESS,
        FDSOURCEADDRESS, MAXIMUMBANDWIDTHPERJOB, PKIKEYPAIR, PKIMASTERKEY,
        NOTES,
        ]
    BOOL_KEYS = [AUTOPRUNE, PKIENCRYPTION, PKISIGNATURES]
    INT_KEYS = [(FDPORT, 9102), (MAXIMUMCONCURRENTJOBS, 1), PRIORITY]
    table = CLIENTS
    # This is kind of a hack used for associating Messages with different
    # resources that are otherwise identically named/numbered.
    IDTAG = 2                   

    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import quotedString, restOfLine, Keyword, nestedExpr, OneOrMore, Word, Literal, removeQuotes, nums, replaceWith, printables
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def _handle_ip(*x):
            a,b,c =  x[2]
            return '  %s = { %s }' % (a,c[0])

        def _handle_fdaddr(*x):
            a,b,c =  x[2]
            self._set(FDADDRESSES, '  %s' % '\n  '.join(c))
            return

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        gr_line = gr_line | np((ADDRESS,), action=self._parse_setter(ADDRESS))
        gr_line = gr_line | np((CATALOG,), action=self._parse_setter(CATALOG_ID, dereference=True))
        gr_line = gr_line | np((PASSWORD,), action=lambda x: x) # Discard the password here!
        gr_line = gr_line | np(PList('file retention'), action=self._parse_setter(FILERETENTION))
        gr_line = gr_line | np(PList('job retention'), action=self._parse_setter(JOBRETENTION))
        gr_line = gr_line | np((PRIORITY,), gr_number, action=self._parse_setter(PRIORITY))
        gr_line = gr_line | np(PList('working directory'), action=self._parse_setter(WORKINGDIRECTORY))
        gr_line = gr_line | np(PList('pid directory'), action=self._parse_setter(PIDDIRECTORY))
        gr_line = gr_line | np(PList('heart beat interval'), action=self._parse_setter(HEARTBEATINTERVAL))
        gr_line = gr_line | np(PList('fd address'), action=self._parse_setter(FDADDRESS))
        gr_line = gr_line | np(PList('fd source address'), action=self._parse_setter(FDSOURCEADDRESS))
        gr_line = gr_line | np(PList('pki key pair'), action=self._parse_setter(PKIKEYPAIR))
        gr_line = gr_line | np(PList('pki master key'), action=self._parse_setter(PKIMASTERKEY))
        gr_line = gr_line | np(PList('fd port'), gr_number, action=self._parse_setter(FDPORT))
        gr_line = gr_line | np(PList('auto prune'), gr_yn, action=self._parse_setter(AUTOPRUNE))
        gr_line = gr_line | np(PList('maximum concurrent jobs'), gr_number, action=self._parse_setter(MAXIMUMCONCURRENTJOBS))
        gr_line = gr_line | np(PList('pki encryption'), gr_yn, action=self._parse_setter(PKIENCRYPTION))
        gr_line = gr_line | np(PList('pki signatures'), gr_yn, action=self._parse_setter(PKISIGNATURES))

        # This is a complicated one
        da_addr = np(('Addr','Port'), Word(printables), lambda x,y,z: ' '.join(z))
        da_ip = np(('IPv4','IPv6','IP'), nestedExpr('{','}', OneOrMore(da_addr).setParseAction(lambda x,y,z: ' ; '.join(z)))).setParseAction(_handle_ip)
        da_addresses = np(('fd addresses', FDADDRESSES), nestedExpr('{','}', OneOrMore(da_ip)), _handle_fdaddr)

        gr_res = OneOrMore(gr_line|da_addresses)
        result = gr_res.parseString(string, parseAll=True)
        return 'Client: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''Convert a Client into a string suitable for inclusion into a Director's configuration.'''
        self.output = ['Client {\n  Name = "%(name)s"' % self,'}']
        self.output.insert(-1, '  %s = "%s"' % (CATALOG.capitalize(), self._fk_reference(CATALOG_ID)[NAME]))
        if getattr(self,DIRECTOR_ID, None):
            a = PasswordStore(self[ID], self.director_id)
            if getattr(a, PASSWORD, None): self.output.insert(-1,'  Password = "%s"' % a.password)
        for key in [ADDRESS, FDPORT, FILERETENTION,
                    JOBRETENTION, MAXIMUMCONCURRENTJOBS,
                    MAXIMUMBANDWIDTHPERJOB, PRIORITY
                    ]:
            self._simple_phrase(key)
        self._yesno_phrase(AUTOPRUNE)
        return '\n'.join(self.output)

# }}}
    # {{{ fd(): return the string that describes the filedaemon configuration for this Client

    def fd(self):
        '''Generate the Client configuration as appropriate for a file daemon'''
        self.output = ['Client {\n  Name = "%(name)s"' % self, '}']
        
        for key in [WORKINGDIRECTORY, PIDDIRECTORY, HEARTBEATINTERVAL, MAXIMUMCONCURRENTJOBS,
                    FDPORT, FDADDRESS, FDSOURCEADDRESS, SDCONNECTTIMEOUT,
                    MAXIMUMNETWORKBUFFERSIZE, MAXIMUMBANDWIDTHPERJOB, PKIKEYPAIR, PKIMASTERKEY
                    ]:
            self._simple_phrase(key)
        if self[FDADDRESSES]:
            self.output.insert(-1, '  %s {' % FDADDRESSES.capitalize())
            self.output.insert(-1,  self[FDADDRESSES])
            self.output.insert(-1, '  }')
        self._yesno_phrase(PKIENCRYPTION)
        self._yesno_phrase(PKISIGNATURES)
        return '\n'.join(self.output)

    # }}}
    # {{{ _cli_special_setup(): add in password support on the CLI

    def _cli_special_setup(self):
        '''This is *all* for setting the password.  Client passwords are associated
        with Directors, which in turn may or may not be monitors for this
        client, and different for other clients.  Look at the PasswordStore
        object in util.py for further detail.

        '''
        self._cli_parser_group(
            [PASSWORD, (DIRECTOR,None,'Director (name or ID) to associate with a password'), (MONITOR,None, '[yes|no]')],
            "Password set/change",
            "Passwords are associated with Directors, so changing a password requires "
            "that you specify the Director to which that password applies.  Also, Directors may be "
            "restricted to a monitoring role.  Specify a value of 'generate' for an auto-generated password."
            )
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle password parsing

    def _cli_special_do_parse(self, args):
        '''When setting the password, ensure that a Director is referenced.  If
        that's the case, make the appropriate updates.'''
        if (args.password == None) and (args.director == None): return # Nothing to do!
        if (args.password == None) ^ (args.director == None):
            print('\n***WARNING***: You must specify both a password and a director to change a password.  Password not changed.\n')
            return              # Bail on any password changes, but otherwise continue
        d = bacula_tools.Director().search(args.director)
        if not d[ID]:
            print('\n***WARNING***: Unable to find a director using "%s".  Password not changed\n' % args.director)
            return
        password = PasswordStore(self[ID], d[ID])
        password.password = args.password
        if args.monitor:
            if args.monitor.lower() in TRUE_VALUES: password.monitor = 1
            else: password.monitor = 0
        password.store()
        return

# }}}
    # {{{ _cli_special_print(): print out passwords

    def _cli_special_print(self):
        '''Prints out the passwords and the directors with which they are associated'''
        print('\nPasswords:')
        sql = 'select director_id from %s where %s = %%s' % (PasswordStore.table, PasswordStore.column1)
        for row in self.bc.do_sql(sql, self[ID]):
            password = PasswordStore(self[ID], row[0])
            d = bacula_tools.Director().search(row[0])
            retval = '%30s: %s' % (d[NAME], password.password)
            if password.monitor: retval += ' (monitor)'
            print(retval)
        return

    # }}}
    # {{{ _cli_special_clone(oid):

    def _cli_special_clone(self, oid):
        '''When cloning a Client, assume that we want to also clone the passwords'''
        select = 'SELECT %s, director_id, password, monitor FROM client_pwords WHERE client_id = %%s' % self[ID]
        insert = 'INSERT INTO client_pwords (client_id, director_id, password, monitor) VALUES (%s,%s,%s,%s)'
        for row in self.bc.do_sql(select, oid): self.bc.do_sql(insert, row)
        return

# }}}

# Implement the CLI for managing Clients
def main():
    s = Client()
    s.cli()

if __name__ == "__main__": main()



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Catalog(DbDict):
    SETUP_KEYS = [DBADDRESS, DBNAME, DBPORT, DBSOCKET, PASSWORD, USER]
    SPECIAL_KEYS = [DIRECTOR_ID,]
    table = CATALOGS
    # {{{ parse_string(string, director): Entry point for a recursive descent parser

    def parse_string(self, string, director):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums, printables
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        gr_line = gr_line | np((USER, 'dbuser', 'db user'), action=self._parse_setter(USER))
        gr_line = gr_line | np((PASSWORD, 'dbpassword', 'db password'), action=self._parse_setter(PASSWORD))
        gr_line = gr_line | np(PList(DBSOCKET), action=self._parse_setter(DBSOCKET))
        gr_line = gr_line | np(PList(DBPORT), gr_number, action=self._parse_setter(DBPORT))
        gr_line = gr_line | np(PList('db name'), action=self._parse_setter(DBNAME))
        gr_line = gr_line | np(PList('db address'), action=self._parse_setter(DBADDRESS))
        gr_res = OneOrMore(gr_line)

        result = gr_res.parseString(string, parseAll=True)
        self._set(DIRECTOR_ID, director[ID])
        return 'Catalog: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''Convert a Catalog into its string representation.  The result can be
        inserted directly into the configuration of the director.'''
        self.output = ['Catalog {\n  Name = "%(name)s"' % self,'}']
        
        for key in self.SETUP_KEYS:
            self._simple_phrase(key)
        return '\n'.join(self.output)

# }}}
    # {{{ search(key=None):

    def search(self, key=None):
        '''Override the standard self.search() function to look up a catalog by
        Director ID.  This is primarily used when generating the
        configuration for a director, as you'll have a Director ID and need
        to look up the associated catalog.'''
        DbDict.search(self, key)
        if self[ID]: return self
        if not self[DIRECTOR_ID]: return self # can't look myself up if I don't have any attributes
        new_me = self.bc.value_check(self.table, DIRECTOR_ID, self[DIRECTOR_ID], asdict=True)
        try: self.update(new_me[0])
        except Exception as e: pass
        [getattr(self, x)() for x in dir(self) if '_load_' in x]
        return self

    # }}}
    # {{{ _cli_special_setup(): add in Director ID support

    def _cli_special_setup(self):
        self.parser.add_option('--director', help='The name or ID of the Director that uses this catalog')
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle Director ID parsing

    def _cli_special_do_parse(self, args):
        '''It should be noted that a Catalog can be associated with one, and only
        one, Director.  This function handles changing the Director from the CLI.'''
        if args.director == None: return # Nothing to do!
        d = bacula_tools.Director()
        d.search(args.director)
        if not d[ID]: d.search(args.director)
        if not d[ID]:
            print('\n***WARNING***: Unable to find a director using "%s".  Association not changed\n' % args.director)
            return
        self._set(DIRECTOR_ID, d[ID])
        return

# }}}
    # {{{ _cli_special_print(): print out passwords

    def _cli_special_print(self):
        '''De-reference the Director and print its name.'''
        d = bacula_tools.Director().search(self[DIRECTOR_ID])
        fmt = '%' + str(self._maxlen) +'s: %s'
        print(fmt % (DIRECTOR, d[NAME]))
        return

    # }}}

# Implement the CLI for managing Catalogs
def main():
    s = Catalog()
    s.cli()

if __name__ == "__main__": main()



from __future__ import print_function
try: from . import *
except: from bacula_tools import *
import bacula_tools
import re, os, sys, filecmp, optparse, random

# Extra stuff needed for the console/daemon tools
import socket, hmac, base64, hashlib, time
from random import randint
from struct import pack, unpack

# {{{ guess_os():

os_bits = {
    'bsd': 'FreeBSD',
    'BSD': 'FreeBSD',
    'Macintosh': 'OSX',
    'OS X': 'OSX',
    'apple-darwin': 'OSX',
    'msie': 'Windows',
    'MSIE': 'Windows',
    }

def guess_os():
    '''This is a "futures" item to support a not-yet-written CGI.  Knowing the
    OS will, in turn, possibly give you information on where various
    directories default.'''
    agent = os.environ.get('HTTP_USER_AGENT', '')
    environ = os.environ.get('PATH_INFO','')
    for k in os_bits.keys():
        if k in agent or k in environ:
            return os_bits[k]
    return 'Linux'              # Default

# }}}
# {{{ generate_password():

def generate_password(length=44):
    '''Genearte a new password.'''
    possible = 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890'
    return ''.join(random.choice(possible) for _ in xrange(length))

# }}}
# {{{ guess_schedule_and_filesets(hostname, os):

def guess_schedule_and_filesets(hostname, os):
    '''Another "futures" function to support a CGI.  When a host registers
    itself, you may want to have automatic selection of appropriate scheduls
    and filesets.'''
    results = []
    for (variable, matcher, fileset, schedule) in _guessing_rules:
        if matcher.match(locals()[variable]): results.append((fileset, schedule))
    return results or _default_rules

# }}}
# {{{ die(*msg):

def die(*msg):
    '''Print out error messages and exit.'''
    print("An error occurred and %s is no longer able to continue\n" % sys.argv[0])
    print('\n'.join(msg))
    print('\n')
    exit(-1)
    return                  # This will never happen, but it makes auto-formatting a little happier.

# }}}

    
class ConfigFile(object):
    '''Easy config file management wrapper.

    This will give you a file-like object that can be easily used to write
    out various bits of output.  If the new output differs from the current
    contents of the target configuration file, upon close this will
    overwrite the destination file and return True.  This, in turn, can let
    you drive processes like "restart bacula-sd if the config file
    changes".

    '''
    FILEHEADER = "# This config file generated by %s script.\n#\n#DO NOT EDIT THIS FILE BY HAND\n\n" % sys.argv[0]
    # {{{ __init__(filename)    # file to update (or not, as the case may be)

    def __init__(self, filename):
        self.filename = filename
        self.newfilename = filename + '.new'
        self.fh = open(self.newfilename, 'w')
        self.fh.write(self.FILEHEADER)
        debug_print("Opened %s", self.newfilename)
        return

    # }}}
    # {{{ close(*data):              # Returns True if an actual update happens

    def close(self, *data):
        '''Write out the data.  Compare that with the target destination file and
        overwrite if necessary.'''
        if data: self.write(*data)
        self.fh.flush()
        self.fh.close()
        try: test_value = filecmp.cmp(self.filename, self.newfilename)
        except: test_value = False
        if test_value:
            os.unlink(self.newfilename)
            debug_print("\t%s doesn't need to be updated", self.filename)
            return False
        debug_print("\tupdating %s", self.filename)
        os.rename(self.newfilename, self.filename)
        return True

# }}}
    # {{{ write(*data): writes data out to the file, appending newlines for each item passed in

    def write(self, *data):
        '''Write data out, with newlines after each item.'''
        for line in data:
            self.fh.write(str(line))
            self.fh.write('\n')
        return

# }}}

class PasswordStore(object):
    '''Client resources have passwords associated.  This class helps manage
    that.  Instantiate with a Client ID and a Director ID.'''
    bc = bacula_tools.Bacula_Factory()
    column1 = 'client_id'
    table = 'client_pwords'
    _select = 'SELECT * FROM %s where %s = %%s and %s = %%s'
    _insert = 'INSERT INTO %s (%s, director_id, password %s) values (%%s, %%s, %%s %s)'
    _update = 'UPDATE %s set password=%%s %s where %s=%%s and director_id=%%s'
    _delete = 'DELETE FROM %s where %s=%%s and director_id=%%s'
    # {{{ __init__(id1, director_id):

    def __init__(self, id1, director_id):
        self.id = id1
        self.director_id = director_id
        self.load()
        return

        # }}}
    # {{{ load():

    def load(self):
        '''Load data from the database'''
        sql = self._select % (self.table, self.column1, DIRECTOR_ID)
        value = self.bc.do_sql(sql, (self.id, self.director_id), asdict=True)
        if len(value) == 1:
            value = value[0]
            self.password = value[PASSWORD]
            if value.has_key(MONITOR): self.monitor = value[MONITOR]
        return

        # }}}
    # {{{ store():

    def store(self):
        '''Write the data out to the database'''
        if self.password == GENERATE: self.password = generate_password()
        if not self.password:
            sql = self._delete % (self.table, self.column1)
            self.bc.do_sql(sql, (self.id, self.director_id))
            return
        if hasattr(self, MONITOR):
            values = (self.password, self.monitor)
            m = ", monitor=%s"
            n = ', monitor'
            i = ', %s'
        else:
            values = (self.password,)
            m = ""
            n = ""
            i = ''
        try:
            sql = self._insert % (self.table, self.column1, n, i)
            self.bc.do_sql(sql, (self.id, self.director_id) + values)
        except Exception as e:
            self.bc.do_sql(self._update % (self.table, m, self.column1), values + (self.id, self.director_id))
        return

        # }}}

class StoragePasswordStore(PasswordStore):
    '''The Storage daemons have slightly different requirements from Clients.
    Instantiate with a Storage ID and a Director ID.'''
    column1 = 'storage_id'
    table = 'storage_pwords'
    def __str__(self):
        d = bacula_tools.Director().search(self.director_id)
        return '%s: %s' % (d[NAME], self.password)
        
class DbDict(dict):
    '''Base class for all of the things derived from database rows.  It's badly
    overloaded with functionality, but it sure is convenient to do it this way.'''
    brace_re = re.compile(r'\s*(.*?)\s*\{\s*(.*)\s*\}\s*', re.MULTILINE|re.DOTALL)
    name_re = re.compile(r'^\s*name\s*=\s*(.*)', re.MULTILINE|re.IGNORECASE)
    bc = bacula_tools.Bacula_Factory()
    output = []
    SETUP_KEYS = []
    INT_KEYS = []
    BOOL_KEYS = []
    prefix = '  '               # Used for spacing out members when printing
    table = 'override me'       # This needs to be overridden in every subclass, before calling __init__
    IDTAG = 0                   # Only used for director/client/storage objects
    # {{{ __init__(row={}, string=None, **kwargs): pass in a row (as a dict)
    def __init__(self, row={}, string = None, **kwargs):
        '''Sets up instance variables and initializes the key/value pairs.
        If a row is passed in, update they keystore with it.
        If a string is passed in, it will be parsed via pyparsing.
        kwargs is *also* used for updating key/value pairs.'''
        dict.__init__(self)
        self.parser = None
        self.special = None
        self[ID] = None         # Ensure we have an ID
        self[NAME] = None       # Ensure we have an NAME
        # This allows flexibility in key setup/declaration, which in turn
        # will allow intelligent groupings to make parse/set/get code
        # somewhat simpler (or at least more clear).
        for x in dir(self):
            if not '_KEYS' in x: continue # Look only for, e.g., NULL_KEYS, SETUP_KEYS, TRUE_KEYS, etc
            for key in getattr(self, x):  # If it's a simple value,
                if type(key) == str: self[key] = None # Assign none
                else: self[key[0]] = key[1]           # Otherwise assume [1] is the desired default  value
        self.update(row)
        if string: self.parse_string(string)
        for key in kwargs: setattr(self, key, kwargs[key])
        self.word = self.table
        if self.word[-1] == 's': self.word = self.word[:-1]
        return

    # }}}
    # {{{ search(key=None):

    def search(self, key=None):
        '''Search for ourself using a number of different methods.  If no key is
        passed in, try self[NAME] and then self[ID].  Otherwise, try the
        passed-in key as an ID if it can be converted to an INT, otherwise
        assume it's a name.  In the event there is no proper row found,
        return a suggestion as to something better to search for, if
        possible.'''
        if not key:
            debug_print('DbDict.search: no key, checking for NAME and ID: table "%s", name "%s", id "%s"'% (self.table, self[NAME], self[ID]))
            if self[NAME]: new_me = self.bc.value_check(self.table, NAME, self[NAME], asdict=True)
            elif self[ID]: new_me = self.bc.value_check(self.table, ID, self[ID], asdict=True)
        else:
            if type(key) == list or type(key) == tuple: key=key[0]
            try: new_me = self.bc.value_check(self.table, ID, int(key), asdict=True)
            except: new_me = self.bc.value_check(self.table, NAME, key, asdict=True)
        try: self.update(new_me[0])
        except Exception as e: pass
        if self[ID]: [getattr(self, x)() for x in dir(self) if '_load_' in x]
        return self

    # }}}
    # {{{ delete(): delete it from the database

    def delete(self):
        '''Delete itself from the database.'''
        self.bc.do_sql('DELETE FROM %s WHERE id = %%s' % self.table, self[ID])
        return

# }}}
    # {{{ _set(field, value, boolean=False, dereference=False): handy shortcut for setting and saving values

    def _set(self, field, value, boolean=False, dereference=False):
        '''Instead of overriding the standard setter, I chose to add a separate
        function that will set and save.  We also do some special processing of
        boolean values.'''
        debug_print('setting %s to %s, boolean=%s, dereference=%s', field, value, boolean, dereference)
        if boolean and value != None:
            if value in TRUE_VALUES: value = 1
            elif value in FALSE_VALUES: value = 0
            else:
                print('"%s" is not a proper boolean value: %s not changed' % (value, field))
                return
        if dereference: value = self._fk_reference(field, value)[ID]
        if field == PASSWORD and value == GENERATE: value = bacula_tools.generate_password()
        self[field] = value
        return self._save()

    # }}}
    # {{{ _save(): Save the top-level fileset record

    def _save(self):
        '''Update the database with our data.'''
        if self[ID]:
            keys = [x for x in self.keys() if not x == ID]
            keys.sort()
            sql = 'UPDATE %s SET %s WHERE id = %%s' % (self.table,
                                                       ', '.join(['`%s` = %%s' % x for x in keys]))
            values = tuple([self[x] for x in keys] + [self[ID],])
            return self.bc.do_sql(sql, values)
        sql = 'INSERT INTO %s (`%s`) VALUES (%s)' % (self.table, '`,`'.join(self.keys()), ','.join(['%s' for x in self.keys()]))
        debug_print(sql, self.values())
        try:
            self.bc.do_sql(sql, tuple(self.values()))
            return self.search()
        except Exception as e:
            if e.args[0] == 1062: die('\t%s "%s" already exists.  You must delete it first.' % (self.word.capitalize(), self[NAME]))
            print(e)
            raise

# }}}
    # {{{ _set_name(name): set my name

    def _set_name(self, name):
        '''Search the database for an existing element associated with name and
        load it up.  Also, run any _load_ hooks that exist.

        '''
        row = self.bc.value_ensure(self.table, NAME, name.strip(), asdict=True)[0]
        self.update(row)
        [getattr(self, x)() for x in dir(self) if '_load_' in x]
        return

    # }}}
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        We're cheating and treating this object as a blob.  Also, the
        *_config parameters will potentially be used to tell us to discard
        passwords.  Sigh.
        '''
        g = self.name_re.search(string).groups()
        self._set_name(g[0].strip())
        string = self.name_re.sub('', string)
        data = '\n  '.join([x.strip() for x in string.split('\n') if x])
        self._set(DATA, data)
        return "%s: %s" % (self.table.capitalize(), self[NAME])

    # }}}
    # {{{ _parse_setter(key, c_int=False, dereference=False):

    def _parse_setter(self, key, c_int=False, dereference=False):
        '''Shortcut called by parser for setting values'''
        def rv(value):
            if c_int: self._set(key, int(value[2].strip()), dereference=dereference)
            else: self._set(key, value[2].strip(), dereference=dereference)
        return rv

# }}}
    # {{{ _simple_phrase(key, quoted=True):

    def _simple_phrase(self, key, quoted=True):
        '''Shortcut for formatting a simple key/value pair.'''
        if not type(key) == str: key = key[0]
        if self[key] == None: return
        if 'retention' in key: quoted = False
        if 'size' in key: quoted = False
        if 'bytes' in key: quoted = False
        try:
            int(self[key])
            value = self[key]
        except:
            if quoted: value = '"' + self[key] + '"'
            else: value = self[key]
        self.output.insert(-1,'%s%s = %s' % (self.prefix, key.capitalize(), value))
        return

    # }}}
    # {{{ _yesno_phrase(key):

    def _yesno_phrase(self, key):
        '''Formatting shortcut for string representation of booleans.'''
        if not type(key) == str: key = key[0]
        value = self[key]
        if value == None: return
        if value == '0': value = NO
        else: value = YES
        self.output.insert(-1,'%s%s = %s' % (self.prefix,key.capitalize(), value))
        return

    # }}}
    # {{{ fd(): stub function to make testing a little easier

    def fd(self):
        '''Stub'''
        return ''

    # }}}
    # {{{ _fk_reference(fk, string=None): Set/get fk-references

    def _fk_reference(self, fk, string=None):
        '''Shortcut for vivifying objects related to foreign keys'''
        debug_print('_fk_reference %s: %s, %s' % (fk, string, self[fk]))
        obj = bacula_tools._DISPATCHER[fk.replace('_id','')]()
        if string:
            obj.search(string.strip())
            if not obj[ID]: obj._set_name(string.strip())
            if not self[fk] == obj[ID]: self._set(fk, obj[ID])
        else: obj.search(self[fk])
        return obj

# }}}

    # Perhaps i'm overloading this class too much here, but
    # a) I can never remember how to do mix-ins, and
    # b) the things I want to do all reference internals.
    # {{{ cli(): Instantiate an option parser and add all the standard bits to it

    def cli(self):
        '''Builds and runs a CLI that applies to every DbDict-derived object.'''
        self.parser = optparse.OptionParser(description='Manage Bacula %ss.' % self.word,
                                            usage='usage: %%prog [options] [%s]' % self.word)
        self.parser.add_option('--create', action='store_true',
                               default=False, help='Create the given %s' % self.word)
        self.parser.add_option('--delete', action='store_true',
                               default=False, help='Delete the given %s' % self.word)
        self.parser.add_option('--rename', metavar='NEW_NAME',
                               help='Rename the given %s' % self.word)
        self.parser.add_option('--list', action='store_true',
                               default=False, help='List the available %ss' % self.word)
        self.parser.add_option('--clone', metavar='CLONE_NAME',
                               help='Duplicate the given %s' % self.word)

        # Add in various groups of things to set
        self._cli_parser_group(self.BOOL_KEYS, "Boolean Setters",
                               "Set with 0/1/yes/no/true/false or '' (empty string) to unset.",
                               metavar = '[yes|no]'
        )

        self._cli_parser_group(self.INT_KEYS, "Integer Setters",
                               "These accept integers or '' (empty string) to unset.", type='int', metavar='number')
        
        self._cli_parser_group(self.SETUP_KEYS, "Setters",
                               "These options are used for setting various values.  "
                               "Use an empty string, e.g. '' to unset the value for strings variables.  "
                               "You should be aware that no sanity checking is done here, so it is quite possible "
                               "to break your configuration while using them.  Caveat emptor."
                           )
        self._cli_special_setup()
        return self._cli_do_parse()

        # }}}
    # {{{ _cli_do_parse(): actually parse the CLI and act on it

    def _cli_do_parse(self):
        '''Parse the standard CLI options.'''
        (args, client_arg) = self.parser.parse_args()


        if args.delete and (args.create or args.rename or args.clone):
            die('','If you delete then there is no sense in doing anything else.',
                "You didn't think this one out very well, did you?.")

        if args.list:
            for row in self.bc.do_sql('select name from %s order by name' % self.table): print(row[0])
            if not client_arg: exit()

        if not client_arg:
            self.parser.print_help()
            exit()

        name_or_num = client_arg[0]
        if args.create: self._set_name(name_or_num)

        self.search(name_or_num)
        
        if not self[ID]:
            print('No such %s: %s' % (self.word, name_or_num))
            exit()

        if args.delete:
            self.delete()
            print('Deleted %s' % name_or_num)
            exit()                      # Nothing to print out now

        if args.rename: self._set(NAME, args.rename)

        if args.clone:
            oid = self[ID]
            self[ID] = None
            self[NAME] = args.clone
            self._save()
            self._cli_special_clone(oid)

        self._cli_option_processor(args, self.BOOL_KEYS, boolean=True)
        self._cli_option_processor(args, self.INT_KEYS)
        self._cli_option_processor(args, self.SETUP_KEYS)
        self._cli_special_do_parse(args)

        self._cli_printer()
        return

        # }}}
    # {{{ _cli_printer():

    def _cli_printer(self):
        '''Print the object out on the CLI'''
        maxlen = 10
        keylist = []
        for key in self.BOOL_KEYS + self.INT_KEYS + self.SETUP_KEYS:
            if not type(key) == str: key = key[0]
            keylist.append(key)
        keylist.sort()
        for key in keylist:
            if len(key) > maxlen: maxlen = len(key)
        maxlen += 4
        self._maxlen = maxlen
        fmt = '%' + str(maxlen) + 's: %s'
        print(fmt % ('ID', str(self[ID])))
        try: print(fmt % ('NAME', str(self[NAME])))
        except: pass            # One child class doesn't have a NAME
        for key in keylist: print(fmt % (key, str(self[key])))
        self._cli_special_print()
        return

    # }}}
    # {{{ _cli_parser_group(keys, label, help_message, **kwargs): add an option group to the CLI parser

    def _cli_parser_group(self, keys, label, help_message, **kwargs):
        '''Shortcut for creating a group of options.'''
        if not keys: return
        group = optparse.OptionGroup(self.parser, label, help_message)
        for key in keys:
            if not type(key) == str:
                if len(key) > 2: group.add_option('--' + key[0], help=key[2], **kwargs)
                else: group.add_option('--' + key[0], **kwargs)
            else: group.add_option('--' + key, **kwargs)
        self.parser.add_option_group(group)

        # }}}
    # {{{ _cli_option_processor(args, items, boolean=False, dereference=False):

    def _cli_option_processor(self, args, items, boolean=False, dereference=False):
        '''Shortcut to handle CLI options in a standard and easy way.'''
        for key in items:
            if not type(key) == str: key = key[0]
            value = getattr(args, key)
            if value == None: continue
            if value == '': value = None
            try: self._set(key, value, boolean, dereference)
            except: pass
        return

        # }}}
        
    # these empty functions provide placeholders for subclasses to call to
    # enable more complicated behaviors.
    def _cli_special_setup(self): pass
    def _cli_special_do_parse(self, args): pass
    def _cli_special_clone(self, oid): pass
    def _cli_special_print(self): pass

class PList(list):
    '''This bizarre construct takes a phrase and lazily turns it into a list
    that is all the permutations of the phrase with all spaces removed.
    Further, this list is sorted such that the first element is the
    original phrase, while the last one has no spaces at all.  It's kind of
    a weird thing, but it makes the string parsing declarations much, much
    more compact and efficient.

    '''
    # {{{ __init__(phrase):

    def __init__(self, phrase):
        list.__init__(self)
        result = self._p2(phrase.split(' '))
        result.sort()
        self.extend(result)
        return

    # }}}
    # {{{ _p2(ary):

    def _p2(self, ary):
        '''Recursive permutation function that returns either single string or a
        pair of strings, one of which is null-joined and the other of which is
        space-joined.'''
        if len(ary) == 1: return ary
        if len(ary) == 2: return [''.join(ary), ' '.join(ary)]
        results = []
        for x in self._p2(ary[1:]):
            results.append(ary[0] + x)
            results.append(ary[0] + ' ' + x)
        return results

    # }}}
    

class BSock:
    '''Sometimes, you want to talk to various Bacula daemons without the
    overhead of firing up bconsole, particularly since that will involve
    shell interaction as well fun parsing foo.  This will make it a bit
    easier, not to mention making dealing with timeouts a lot more
    mangeable.
    '''
# {{{ __init__(address, password, myname, port, debug=False, timeout=5):
    def __init__(self, address, password, myname, port, debug=False, timeout=5):
        '''Address, password, myname, and port are all mandatory.

        address = the destination with which you want to communicate. (None -> 127.0.0.1)
        myname = the "name" with which a password is associated.  This is not as obvious as you might hope.

        '''
        if not address: address = '127.0.0.1'
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection.settimeout(timeout) # Don't take forever trying to do stuff
        self.log('connecting to: %s' % str((address,port)))
        self.connection.connect((address,port))
        self.password = password
        self.name = myname
        self.DEBUG = debug
        return
# }}}
# {{{ log(msg): Debugging output

    def log(self, msg):
        '''Write out a single message to stderr, IFF self.DEBUG.'''
        if self.DEBUG:
            sys.stderr.write(msg)
            sys.stderr.flush()
        return

    # }}}
# {{{ auth(): authenticate a new connection

    def auth(self):
        '''Authenticated this client with the target service.

        I should note that, as written, the target service is *not* mutually authenticated.  

        The bulk of this was written by Matthew Ife, so thanks!'''
        
        self.send("Hello %s calling\n" % (self.name,)) # this is effectively our username
        challenge = self.recv() # Receive the challenge response
        m = re.search("auth cram-md5 (<.+?>)", challenge) # parse the challenge out of the returned string
        chal = m.group(1)
        
        pw = hashlib.md5(self.password).hexdigest()
        self.send(base64.b64encode(hmac.new(pw, chal).digest())[:-2]) # hmac and base64 encode the request

        result = self.recv() # receive response
        if result != "1000 OK auth\n": raise ValueError("Authentication Failed %s" % (result,))# failed

        # send our challenge response
        self.send("auth cram-md5 <%d.%d@%s> ssl=0\n" % (randint(1,99999999), int(time.time()), self.name))
        self.recv()                 # get the response back
        self.send("1000 OK auth\n") # Dont even check the response here!

        # This is basically cheating the protocol spec! :-)
        data = self.recv()
        if not re.match(".* OK.*",data): # auth complete
            raise ValueError("Unexpected packet received %s" % (data,))
        self.auth = True
        return data

    # }}}
# {{{ send(message): send a message to the remote daemon

    def send(self, message):
        '''Send a properly encoded messages to the connected service.'''
        ldata = pack('!i',len(message))
        self.connection.send(ldata)
        self.log( 'sending:  (%d) %s\n' % (len(message), message))
        self.connection.send(message)
        return

    # }}}
# {{{ recv(): Get a one line response from the remote daemon

    def recv(self):
        '''Read a (theoretically) single-line response from the connected service.'''
        msglen = unpack('!i', self.connection.recv(4))[0]
        if msglen < 0: return ''
        response = self.connection.recv(msglen)
        self.log( 'received: %s' % response)
        return response

    # }}}
# {{{ recv_all(): receive a multi-line response from the remote daemon

    def recv_all(self):
        """Gets all lines of a request"""
        r = ""
        s = self.recv()
        while s:
            r += s
            s = self.recv()
        return r

    # }}}
# {{{ version(): request version info from the remote daemon (useless?)

    def version(self):
        '''Request the version string from the connected services.'''
        self.send('version')
        return self.recv()

    # }}}
# {{{ status(args=''): request status of the remote daemon

    def status(self, args=''):
        '''Ask the connected service for its status.'''
        if args: self.send('.status %s' % args)
        else: self.send('status')
        return self.recv_all()

    # }}}
# {{{ _time(): format the time for uniqueifying various things

    def _time(self):
        '''Format the time for uniqueifying various things'''
        return time.strftime('%F_%H.%M.%S_00')

    # }}}



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Storage(DbDict):
    SETUP_KEYS = [SDPORT, ADDRESS, DEVICE, MEDIATYPE, MAXIMUMCONCURRENTJOBS, HEARTBEATINTERVAL,
                 WORKINGDIRECTORY, PIDDIRECTORY, CLIENTCONNECTWAIT, SDADDRESSES]
    BOOL_KEYS = [AUTOCHANGER, ALLOWCOMPRESSION]
    table = STORAGE
    # This is kind of a hack used for associating Messages with different
    # resources that are otherwise identically named/numbered.
    IDTAG = 3
    dir_keys = [SDPORT, ADDRESS, DEVICE, MEDIATYPE, MAXIMUMCONCURRENTJOBS, HEARTBEATINTERVAL]
    sd_keys = [WORKINGDIRECTORY, PIDDIRECTORY, CLIENTCONNECTWAIT, SDADDRESSES,
               SDPORT, MAXIMUMCONCURRENTJOBS, HEARTBEATINTERVAL]
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums, printables
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def _handle_ip(*x):
            a,b,c =  x[2]
            return '  %s = { %s }' % (a,c[0])

        def _handle_addr(*x):
            a,b,c =  x[2]
            self._set(SDADDRESSES, '  %s' % '\n  '.join(c))
            return

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        gr_line = gr_line | np(PList('sd port'), gr_number, action=self._parse_setter(SDPORT))
        gr_line = gr_line | np((ADDRESS,'sd address', SDADDRESS), action=self._parse_setter(ADDRESS))
        gr_line = gr_line | np((PASSWORD,), action=lambda x: x)
        gr_line = gr_line | np((DEVICE,), action=self._parse_setter(DEVICE))
        gr_line = gr_line | np(PList('media type'), action=self._parse_setter(MEDIATYPE))
        gr_line = gr_line | np(PList('auto changer'), gr_yn, action=self._parse_setter(AUTOCHANGER))
        gr_line = gr_line | np(PList('maximum concurrent jobs'), gr_number, action=self._parse_setter(MAXIMUMCONCURRENTJOBS))
        gr_line = gr_line | np(PList('allow compression'), gr_yn, action=self._parse_setter(ALLOWCOMPRESSION))
        gr_line = gr_line | np(PList('heartbeat interval'), action=self._parse_setter(HEARTBEATINTERVAL))

        gr_line = gr_line | np(PList('working directory'), action=self._parse_setter(WORKINGDIRECTORY))
        gr_line = gr_line | np(PList('pid directory'), action=self._parse_setter(PIDDIRECTORY))
        gr_line = gr_line | np(PList('client connect wait'), action=self._parse_setter(CLIENTCONNECTWAIT))

        # This is a complicated one
        da_addr = np(('Addr','Port'), Word(printables), lambda x,y,z: ' '.join(z))
        da_ip = np(('IPv4','IPv6','IP'), nestedExpr('{','}', OneOrMore(da_addr).setParseAction(lambda x,y,z: ' ; '.join(z)))).setParseAction(_handle_ip)
        da_addresses = np(PList('sd addresses'), nestedExpr('{','}', OneOrMore(da_ip)), _handle_addr)


        gr_res = OneOrMore(gr_line)
        result = gr_res.parseString(string, parseAll=True)
        return 'Storage: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Storage suitable for inclusion in a Director configuration.'''
        self.output = ['Storage {\n  Name = "%(name)s"' % self,'}']
        if self.director_id:
            a = StoragePasswordStore(self[ID], self.director_id)
            if getattr(a,PASSWORD,None): self.output.insert(-1,'  Password = "%s"' % a.password)
        
        for key in self.dir_keys: self._simple_phrase(key)
        for key in self.BOOL_KEYS: self._simple_phrase(key)
        return '\n'.join(self.output)

# }}}
    # {{{ sd(): 

    def sd(self):
        '''String representation of a Storage suitable for inclusion in a Storage configuration.'''
        self.output = ['Storage {\n  Name = "%(name)s"' % self,'}']
        for key in self.sd_keys: self._simple_phrase(key)
        # Special keys
        if self[ADDRESS]: self.output.insert(-1,'%sSDAddress = %s' % (self.prefix, '"' + self[ADDRESS] + '"'))
        return '\n'.join(self.output)

# }}}
    # {{{ _cli_special_setup(): add password support

    def _cli_special_setup(self):
        '''Add CLI switches for password handling.'''
        self._cli_parser_group(
            [PASSWORD, (DIRECTOR,None,'Director (name or ID) to associate with a password')],
            "Password set/change",
            "Passwords are associated with directors, so changing a password requires "
            "that you specify the director to which that password applies.  Set the password to "
            "a value of 'generate' to auto-generate a password."
            )
        return

    # }}}
    # {{{ _cli_special_do_parse(args): add password support

    def _cli_special_do_parse(self, args):
        '''Handle CLI switches for password management.'''
        if (args.password == None) and (args.director == None): return # Nothing to do!
        if (args.password == None) ^ (args.director == None):
            print('\n***WARNING***: You must specify both a password and a director to change a password.  Password not changed.\n')
            return              # Bail on any password changes, but otherwise continue
        d = bacula_tools.Director()
        try: d.search(args.director)
        except: d.search(args.director)
        if not d[ID]:
            print('\n***WARNING***: Unable to find a director using "%s".  Password not changed\n' % args.director)
            return
        password = StoragePasswordStore(self[ID], d[ID])
        password.password = args.password
        password.store()
        return

    # }}}
    # {{{ _cli_special_print(): add password support

    def _cli_special_print(self):
        '''Print out the passwords for the CLI.'''
        print('\nPasswords:')
        sql = 'select director_id from %s where %s = %%s' % (StoragePasswordStore.table, StoragePasswordStore.column1)
        for row in self.bc.do_sql(sql, self[ID]):
            password = StoragePasswordStore(self[ID], row[0])
            d = bacula_tools.Director().search(row[0])
            print('%30s: %s' % (d[NAME], password.password))
        return

    # }}}
    # {{{ _cli_special_clone(oid):

    def _cli_special_clone(self, oid):
        '''Storage clones should have the same sets of passwords as the source object.'''
        select = 'SELECT %s, director_id, password FROM storage_pwords WHERE storage_id = %%s' % self[ID]
        insert = 'INSERT INTO storage_pwords (storage_id, director_id, password) VALUES (%s,%s,%s)'
        for row in self.bc.do_sql(select, oid): self.bc.do_sql(insert, row)
        return

# }}}

def main():
    s = Storage()
    s.cli()

if __name__ == "__main__": main()




from . import *
# client for interacting with a file daemon

class FDaemon(BSock):
    '''Client for communicating directly with a file daemon.
    '''
# {{{ __init__(address, password, myname, port=BACULA_FD_PORT, debug=False, timeout=5):
    def __init__(self, address, password, myname, port=BACULA_FD_PORT, debug=False, timeout=5):
        BSock.__init__(self, address, password, 'Director ' + myname,
                                             port, debug, timeout)
        return
# }}}
# {{{ version(): request version info from the remote daemon (useless?)

    def version(self):
        for line in self.status().split('\n'):
            if 'Version' in line: return line
        return 'Unable to determine version'

    # }}}



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Director(DbDict):
    SETUP_KEYS = [ADDRESS, 
                 FD_CONNECT_TIMEOUT, HEARTBEATINTERVAL,
                 PASSWORD, PIDDIRECTORY, QUERYFILE,
                 SCRIPTS_DIRECTORY, SD_CONNECT_TIMEOUT, SOURCEADDRESS, STATISTICS_RETENTION,
                 WORKINGDIRECTORY, DIRADDRESSES]
    INT_KEYS = [(DIRPORT, 9101), MAXIMUMCONCURRENTJOBS, MAXIMUMCONSOLECONNECTIONS,]
    NULL_KEYS = [MESSAGES_ID, ]
    table = DIRECTORS
    # This is kind of a hack used for associating Messages with different
    # resources that are otherwise identically named/numbered.
    IDTAG = 1
    # {{{ parse_string(string, director_config, obj): Entry point for a recursive descent parser

    def parse_string(self, string, director_config, obj):
        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        FTR: this is hideous.
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums, printables
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def _handle_password(*x):
            a,b,c =  x[2]
            if type(obj) == bacula_tools.Client: klass = PasswordStore
            elif type(obj) == bacula_tools.Storage: klass = StoragePasswordStore
            else: print("WTH is a", type(obj), "being passed in?", director_config)
            a = klass(obj[ID], self[ID])
            a.password = c
            a.store()
            return

        def _handle_monitor(*x):
            a,b,c =  x[2]
            a = PasswordStore(obj[ID], self[ID])
            a.monitor = c
            a.store()
            return

        def _handle_ip(*x):
            a,b,c =  x[2]
            return '  %s = { %s }' % (a,c[0])

        def _handle_diraddr(*x):
            a,b,c =  x[2]
            self._set(DIRADDRESSES, '  %s' % '\n  '.join(c))
            return

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p
            
        gr_name = np((NAME,), action=lambda x: self._set_name(x[2]))
        gr_address = np((ADDRESS,), action=self._parse_setter(ADDRESS))
        gr_fd_conn = np(PList('fd connect timeout'), gr_number, self._parse_setter(FD_CONNECT_TIMEOUT, True))
        gr_heart = np(PList('heartbeat interval'), gr_number, self._parse_setter(HEARTBEATINTERVAL, True))
        gr_max_con = np(PList('maximum console connections'),
                        gr_number, self._parse_setter(MAXIMUMCONSOLECONNECTIONS, True))
        gr_max_jobs = np(PList('maximum concurrent jobs'), gr_number, action=self._parse_setter(MAXIMUMCONCURRENTJOBS, True))
        gr_pid = np(PList('pid directory'), action=self._parse_setter(PIDDIRECTORY))
        gr_query = np(PList('query file'), action=self._parse_setter(QUERYFILE))
        gr_scripts = np(PList('scripts directory'), action=self._parse_setter(SCRIPTS_DIRECTORY))
        gr_sd_conn = np(PList('sd connect timeout'), gr_number, self._parse_setter(SD_CONNECT_TIMEOUT, True))
        gr_source = np(PList('source address'), action=self._parse_setter(SOURCEADDRESS))
        gr_stats = np(PList('statistics retention'), action=self._parse_setter(STATISTICS_RETENTION))
        gr_messages = np((MESSAGES,), action=self._parse_setter(MESSAGES_ID, dereference=True))
        gr_work_dir = np(PList('working directory'), action=self._parse_setter(WORKINGDIRECTORY))
        gr_port = np(PList('dir port'), gr_number, self._parse_setter(DIRPORT, True))
        gr_monitor = np((MONITOR,), gr_yn, action=_handle_monitor)

        # This is a complicated one
        da_addr = np(('Addr','Port'), Word(printables), lambda x,y,z: ' '.join(z))
        da_ip = np(('IPv4','IPv6','IP'), nestedExpr('{','}', OneOrMore(da_addr).setParseAction(lambda x,y,z: ' ; '.join(z)))).setParseAction(_handle_ip)
        da_addresses = np(PList('dir addresses'), nestedExpr('{','}', OneOrMore(da_ip)), _handle_diraddr)

        # if this isn't a director, then we ignore the password
        if director_config: gr_pass = np((PASSWORD,), action=self._parse_setter(PASSWORD))
        else: gr_pass = np((PASSWORD,), action=_handle_password)

        gr_res = OneOrMore(gr_name | gr_address | gr_fd_conn | gr_heart | gr_max_con | gr_max_jobs | gr_pass | gr_pid | gr_query | gr_scripts | gr_sd_conn | gr_source | gr_stats | gr_messages | gr_work_dir | gr_port | gr_monitor | da_addresses)

        result = gr_res.parseString(string, parseAll=True)
        return 'Director: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Director, suitable for inclusion in director.conf'''
        self.output = ['Director {\n  Name = "%(name)s"' % self,'}']
        for key in self.NULL_KEYS: self._simple_phrase(key)
        # set the messages
        m = self._fk_reference(MESSAGES_ID)
        self.output.insert(-1, '  %s = "%s"' % (MESSAGES.capitalize(), m[NAME]))
        if self[DIRADDRESSES]:
            self.output.insert(-1, '  %s {' % DIRADDRESSES.capitalize())
            self.output.insert(-1,  self[DIRADDRESSES])
            self.output.insert(-1, '  }')
        return '\n'.join(self.output)

# }}}
    # {{{ fd(): return the string that describes the filedaemon configuration

    def fd(self):
        '''This is what we'll call to dump out the config for the file daemon'''
        self.output = ['Director {\n  Name = "%(name)s"' % self, '}']
        if getattr(self,CLIENT_ID, None):
            a = PasswordStore(self.client_id, self[ID])
            if getattr(a,PASSWORD, None):
                self.output.insert(-1,'  Password = "%s"' % a.password)
                if a.monitor: self.output.insert(-1,'  Monitor = "yes"' )
        return '\n'.join(self.output)

    # }}}
    # {{{ sd(): return the string that describes the storagedaemon configuration

    def sd(self):
        '''This is what we'll call to dump out the config for the storage daemon'''
        self.output = ['Director {\n  Name = "%(name)s"' % self, '}']
        if getattr(self,STORAGE_ID, None):
            a = StoragePasswordStore(self.storage_id, self[ID])
            if getattr(a,PASSWORD, None):
                self.output.insert(-1,'  Password = "%s"' % a.password)
        return '\n'.join(self.output)

    # }}}
    # {{{ _cli_special_setup(): setup the weird phrases that go with directors

    def _cli_special_setup(self):
        '''Handle setting by the CLI of the Message to be used by this Director.'''
        group = optparse.OptionGroup(self.parser,
                                     "Object Setters",
                                     "Various objects associated with a Job")
        group.add_option('--message-set')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle the weird phrases that go with directors

    def _cli_special_do_parse(self, args):
        '''Enable the CLI to set the Message to be used by this Director.'''
        if args.message_set == None: return
        if args.message_set =='': return self._set(MESSAGES_ID, None)
        target = Messages().search(args.message_set)
        if target[ID]: self._set(MESSAGES_ID, target[ID])
        else: print('Unable to find a match for %s, continuing' % args.message_set)
        return

# }}}
    # {{{ _cli_special_print(): print out the weird phrases that go with directors

    def _cli_special_print(self):
        '''Print out the Message for this Director'''
        if not self[MESSAGES_ID]: return
        fmt = '%' + str(self._maxlen) + 's: %s'
        print(fmt % ('Messages', self._fk_reference(MESSAGES_ID)[NAME]))
        return
    # }}}

def main():
    s = Director()
    s.cli()

if __name__ == "__main__": main()



from . import *
# client for interacting with a the Bacula director

class BDirector(BSock):
    '''Client for communicating directly with a director.
    '''
# {{{ __init__(director_object, timeout=5):
    def __init__(self, director_object, debug=False, timeout=5):
        BSock.__init__(self, director_object[ADDRESS], director_object[PASSWORD], '*UserAgent*',
                       int(director_object[DIRPORT]), timeout)
        return
# }}}



from __future__ import print_function
import re
from . import *
import bacula_tools

'''It is the long-term plan that the parser support in the individual
classes will be (somehow) extracted into here as there is a lot of
repetitive boilerplate, and that offends me.

'''

class StringParseSupport:
    '''Parse a string out into top-level resource items and pass them off to the relevant classes.'''
    RB = '}'
    LB = '{'
    file_re = re.compile(r'\s@(.*)\s+', re.MULTILINE)
    comment_re = re.compile(r'#.*', re.MULTILINE)
    semicolon_re = re.compile(r';', re.MULTILINE)
    blankline_re = re.compile(r'^\s+$', re.MULTILINE)
    # {{{ __init__(output):

    def __init__(self, output):
        '''Initialize the instance variables, and set the output device.  There
        should probably be a default set here.'''
        self.output = output
        self.parse_queue = {}
        self.parsed = []
        self.director_config = False
        self.sd_config = False
        self.fd_config = False
        return

        # }}}
    # {{{ file_replacement(string): Replace file import statements with the contents of the files.

    def file_replacement(self, string):
        ''' Do a quick pass through the string looking for file imports.  If/When
        you find any, replace the file import statement with the contents of
        the file to be imported.  Repeat until there are no more file import statements.'''

        # Strip the comments and blank lines out
        string = self.blankline_re.sub('', self.comment_re.sub('', string))

        # Search for any file import statements
        groups = self.file_re.search(string)

        while groups:           # Iterate through the file imports
            filename = groups.group(1)
            string_without_comments = self.comment_re.sub('', open(filename).read())
            string = string.replace(groups.group(), string_without_comments)
            # re-strip blank lines *after* doing the insert.
            string = self.blankline_re.sub('', string)
            # Look again for import statements
            groups = self.file_re.search(string)
        return string

        # }}}
    # {{{ break_into_stanzas(string): segment the configuration data in individual resources

    def break_into_stanzas(self, string):
        '''Split up the input string into resources, and drop each resource into a
        queue to be individually parsed.  The complication is that there
        are nested {} groups, so we have to do some counting and
        re-assembly so we don't break things up in too fine-grained a manner.

        '''
        parts = string.split(self.RB)

        # Split it up into parts, stored in self.parse_queue
        while parts:
            current = parts.pop(0)
            while current.count(self.RB) < (current.count(self.LB) - 1): current += self.RB + parts.pop(0)
            try: name, body = current.split(self.LB,1)
            except:
                self.output(current)
                raise
            name = name.strip().lower()
            self.parse_queue.setdefault(name, []).append(body.strip())
            while parts and parts[0] == '\n': del parts[0]
        return

        # }}}
    # {{{ analyze_queue(): Determine what kind of configuration file we're parsing

    def analyze_queue(self):
        '''Determine what kind of configuration file this is, as that affects the
        parsing process greatly.'''
        if CATALOG.lower() in self.parse_queue: self.director_config = True
        elif DEVICE in self.parse_queue: self.sd_config = True
        else: self.fd_config = True
        return

        # }}}
    # {{{ parse_one_stanza_type(key):

    def parse_one_stanza_type(self, key):
        '''Parse all of the stanzas of one type.  e.g. all of the Clients  '''
        if not key in self.parse_queue: return
        # Actually parse something
        for body in self.parse_queue[key]:
            try:
                obj = bacula_tools._DISPATCHER[key]()
                self.parsed.append(obj)
                if key == DIRECTOR: result = obj.parse_string(body, self.director_config, self.parsed[0])
                elif key in [CATALOG.lower(), MESSAGES, DEVICE]: result = obj.parse_string(body, self.parsed[0])
                else: result = obj.parse_string(body)
                self.output(result)
            except Exception as e:
                msg = '%s: Unable to handle %s at this time:\n%s' % (key.capitalize(), e,body.strip())
                self.output(msg)
        del self.parse_queue[key]
        return

        # }}}
    # {{{ parse_it_all(string): Main parser driver

    def parse_it_all(self, string):
        '''Takes a string resulting from reading in a file, and parse it out.
        Think of this as the driver routing for the entire parsing
        process.

        '''
        string = self.file_replacement(string)

        # It should be observed that this statement causes scripts with
        # embedded semi-colons to break parsing.
        string = self.semicolon_re.sub('\n',  string).replace('\\\n','').replace('\n\n', '\n')

        # Fills in self.parse_queue
        self.break_into_stanzas(string)
        self.analyze_queue()

        # Actually parse the various parts
        # it turns out to be important to do these in a specific order.
        if self.sd_config: self.parse_one_stanza_type(STORAGE)
        if self.fd_config: self.parse_one_stanza_type(CLIENT)
        if self.fd_config: self.parse_one_stanza_type(FILEDAEMON)
        if self.director_config:self.parse_one_stanza_type(DIRECTOR)
        for name in self.parse_queue.keys(): self.parse_one_stanza_type(name)

        return self.parsed

        # }}}
    

def setup_for_parsing():
    '''This is for future use.  I had kind of envisioned monkey-patching the
    parsers into the main classes so that I could re-use stuff more easily,
    as well as possibly benefitting from having all the pyparsing code in a
    single place.  I have no idea if it will ever happen.

    '''
    pass

def parser(string, output=print):
    '''This is the primary entry point for the parser.  Call it with a string
    and a function to be called for ouput.'''
    setup_for_parsing()
    p = StringParseSupport(output)
    return p.parse_it_all(string)



'''Bacula configuration database stuff: common routines, credentials, etc'''
# Bacula_Config and BSock classes and various definitions used by Bacula python programs.

import MySQLdb as db
import MySQLdb.cursors
import os, sys

from . import *

_singleton = None
def Bacula_Factory():
    '''Returns a singleton instance of the Bacula_Config class.  This is a
    simple optimization to reduce the number of database connections used.'''
    global _singleton
    if not _singleton: _singleton = Bacula_Config()
    return _singleton

class Bacula_Config:
    '''Class that wraps up lots of dealings with the configuration database.
    You probably should call Bacula_Factory instead of instantiating this
    directly.

    '''
    # {{{ Various class constants

    CONNECTIONS = {}            # array of connections to the database
    CURRENT_CONNECTION = None   # The currently in-use connection

    # }}}
    # {{{ __db_connect(database=MYSQL_DB, user=MYSQL_USER, passwd=MYSQL_PASS, host=MYSQL_HOST):

    def __db_connect(self, database=MYSQL_DB, user=MYSQL_USER, passwd=MYSQL_PASS, host=MYSQL_HOST):
        '''Connect to the database.  Tries to look up an existing connection to use if possible'''
        key = database+user+passwd+host
        if not self.CONNECTIONS.has_key(key): 
            self.CONNECTIONS[key] = db.connect(db = database, user = user, passwd = passwd, host = host)
        self.CURRENT_CONNECTION = self.CONNECTIONS[key]
        self.CURRENT_CONNECTION.autocommit(True)
        return self.CONNECTIONS[key]

    # }}}
    # {{{ get_cursor():

    def get_cursor(self, **kwargs):
        '''Returns a cursor for querying.  Will automatically connect to the database if necessary.'''
        if not self.CURRENT_CONNECTION: self.__db_connect() # Assume default connection stuff
        return self.CURRENT_CONNECTION.cursor(**kwargs)

    # }}}
    # {{{ do_sql(sql, args=None):

    def do_sql(self, sql, args=None, asdict=False):
        '''A general-purpose SQL query function.  It handles acquiring a cursor
        that returns either a list (default) or dictionary as requested,
        performs the sql, and returns the entire resultset.

        You should not use this for extremely large resultsets.'''
        if asdict: cursor = self.get_cursor(cursorclass=db.cursors.DictCursor)
        else: cursor = self.get_cursor()
        cursor.execute(sql, args)
        return cursor.fetchall()

    # }}}
    # {{{ suggest(table, field, value):

    def suggest(self, table, field, value):
        '''This is an effort to be helpful in the case where you have an idea on
        the name of a resource, but you aren't really sure of the *precise*
        name.  This strips off the first and last characters of your name
        string, wraps the results in %%, and then queries the database
        (limiting the result set to 5).'''
        sql = "select %(field)s from %(table)s where %(field)s like %%s limit 5" % locals()
        mangled_value = '%' + value[1:-1] + '%'
        data = self.do_sql(sql, mangled_value)
        if data:                # OK, we got some partial matches
            return "Possible matches for '%s':\n\t" % value + '\n\t'.join([x[0] for x in data if x])
        return "I was unable to find any close matches to '%(value)s'.  Please try harder.\n" % locals();

    # }}}
    # {{{ value_check(table, field, value, suggest=False, asdict=False):

    def value_check(self, table, field, value, suggest=False, asdict=False):
        '''Check the existence of a value in a column.  If it exists, return the
        row as a dictionary, otherwise die with a suggestion of possible
        alternatives to search for.'''
        sql = "SELECT * FROM `%(table)s` where `%(field)s` = %%s" % locals()
        result = self.do_sql(sql, value, asdict);
        if result: return result
        if not suggest: return result
        return die('No such value (%(value)s) in %(table)s' % locals(),
                   '',
                   self.suggest(table, field, value))

    # }}}
    # {{{ value_ensure(table, field, value, asdict=False):

    def value_ensure(self, table, field, value, asdict=False):
        '''Ensure the existence of a value in a column.  Use this to find-or-create
        a resource record.'''
        if not self.value_check(table, field, value, asdict=asdict):
            self.do_sql("INSERT INTO %(table)s (%(field)s) VALUES (%%s)" % locals(), value);
        return self.value_check(table, field, value, asdict=asdict)

    # }}}



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Device(DbDict):
    SETUP_KEYS = [
        ARCHIVEDEVICE, DEVICETYPE, MEDIATYPE, CHANGERDEVICE, CHANGERCOMMAND, ALERTCOMMAND,
        DRIVEINDEX, MAXIMUMCONCURRENTJOBS, MAXIMUMCHANGERWAIT, MAXIMUMREWINDWAIT,
        MAXIMUMOPENWAIT, VOLUMEPOLLINTERVAL, MOUNTPOINT, MOUNTCOMMAND, UNMOUNTCOMMAND,
        MINIMUMBLOCKSIZE, MAXIMUMBLOCKSIZE, MAXIMUMVOLUMESIZE, MAXIMUMFILESIZE,
        MAXIMUMNETWORKBUFFERSIZE, MAXIMUMSPOOLSIZE, MAXIMUMJOBSPOOLSIZE, SPOOLDIRECTORY,
        MAXIMUMPARTSIZE, CLIENTCONNECTWAIT,
        ]
    BOOL_KEYS = [
        AUTOSELECT, REMOVABLEMEDIA, BLOCKCHECKSUM, HARDWAREENDOFMEDIUM,
        FASTFORWARDSPACEFILE, USEMTIOCGET, AUTOMATICMOUNT,
        BACKWARDSPACERECORD, BACKWARDSPACEFILE, FORWARDSPACERECORD,
        FORWARDSPACEFILE, BLOCKPOSITIONING, AUTOCHANGER, ALWAYSOPEN, CLOSEONPOLL,
        RANDOMACCESS, BSFATEOM, TWOEOF, OFFLINEONUNMOUNT, LABELMEDIA,
        ]
    table = DEVICE
    _insert = 'INSERT INTO device_link (device_id, storage_id) values (%s, %s)'
    _delete = 'DELETE FROM device_link where device_id = %s and storage_id = %s'
    _select = 'SELECT storage_id FROM device_link where device_id = %s'
    # {{{ parse_string(string, obj=None): Entry point for a recursive descent parser

    def parse_string(self, string, obj=None):
        # {{{ boilerplate.  Sigh

        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def np(words, fn = gr_opt_quoted_string, action=None):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        # }}}

        gr_line =           np(PList(NAME), action=lambda x: self._set_name(x[2]))
        gr_line = gr_line | np(PList('alert command'), action=self._parse_setter(ALERTCOMMAND))
        gr_line = gr_line | np(PList('archive device'), action=self._parse_setter(ARCHIVEDEVICE))
        gr_line = gr_line | np(PList('changer command'), action=self._parse_setter(CHANGERCOMMAND))
        gr_line = gr_line | np(PList('changer device'), action=self._parse_setter(CHANGERDEVICE))
        gr_line = gr_line | np(PList('client connect wait'), action=self._parse_setter(CLIENTCONNECTWAIT))
        gr_line = gr_line | np(PList('device type'), action=self._parse_setter(DEVICETYPE))
        gr_line = gr_line | np(PList('drive index'), gr_number, action=self._parse_setter(DRIVEINDEX, c_int=True))
        gr_line = gr_line | np(PList('maximum block size'), action=self._parse_setter(MAXIMUMBLOCKSIZE))
        gr_line = gr_line | np(PList('maximum changer wait'), action=self._parse_setter(MAXIMUMCHANGERWAIT))
        gr_line = gr_line | np(PList('maximum concurrent jobs'), gr_number, action=self._parse_setter(MAXIMUMCONCURRENTJOBS, c_int=True))
        gr_line = gr_line | np(PList('maximum file size'), action=self._parse_setter(MAXIMUMFILESIZE))
        gr_line = gr_line | np(PList('maximum job spool size'), action=self._parse_setter(MAXIMUMJOBSPOOLSIZE))
        gr_line = gr_line | np(PList('maximum network buffer size'), action=self._parse_setter(MAXIMUMNETWORKBUFFERSIZE))
        gr_line = gr_line | np(PList('maximum open wait'), action=self._parse_setter(MAXIMUMOPENWAIT))
        gr_line = gr_line | np(PList('maximum part size'), action=self._parse_setter(MAXIMUMPARTSIZE))
        gr_line = gr_line | np(PList('maximum rewind wait'), action=self._parse_setter(MAXIMUMREWINDWAIT))
        gr_line = gr_line | np(PList('maximum spool size'), action=self._parse_setter(MAXIMUMSPOOLSIZE))
        gr_line = gr_line | np(PList('maximum volume size'), action=self._parse_setter(MAXIMUMVOLUMESIZE))
        gr_line = gr_line | np(PList('media type'), action=self._parse_setter(MEDIATYPE))
        gr_line = gr_line | np(PList('minimum block size'), action=self._parse_setter(MINIMUMBLOCKSIZE))
        gr_line = gr_line | np(PList('mount command'), action=self._parse_setter(MOUNTCOMMAND))
        gr_line = gr_line | np(PList('mount point'), action=self._parse_setter(MOUNTPOINT))
        gr_line = gr_line | np(PList('spool directory'), action=self._parse_setter(SPOOLDIRECTORY))
        gr_line = gr_line | np(PList('unmount command'), action=self._parse_setter(UNMOUNTCOMMAND))
        gr_line = gr_line | np(PList('volume poll interval'), action=self._parse_setter(VOLUMEPOLLINTERVAL))

        gr_line = gr_line | np(PList('always open'), gr_yn, action=self._parse_setter(ALWAYSOPEN))
        gr_line = gr_line | np(PList('auto changer'), gr_yn, action=self._parse_setter(AUTOCHANGER))
        gr_line = gr_line | np(PList('auto select'), gr_yn, action=self._parse_setter(AUTOSELECT))
        gr_line = gr_line | np(PList('automatic mount'), gr_yn, action=self._parse_setter(AUTOMATICMOUNT))
        gr_line = gr_line | np(PList('backward space file'), gr_yn, action=self._parse_setter(BACKWARDSPACEFILE))
        gr_line = gr_line | np(PList('backward space record'), gr_yn, action=self._parse_setter(BACKWARDSPACERECORD))
        gr_line = gr_line | np(PList('block check sum'), gr_yn, action=self._parse_setter(BLOCKCHECKSUM))
        gr_line = gr_line | np(PList('block positioning'), gr_yn, action=self._parse_setter(BLOCKPOSITIONING))
        gr_line = gr_line | np(PList('bsf at eom'), gr_yn, action=self._parse_setter(BSFATEOM))
        gr_line = gr_line | np(PList('close on poll'), gr_yn, action=self._parse_setter(CLOSEONPOLL))
        gr_line = gr_line | np(PList('fast forward space file'), gr_yn, action=self._parse_setter(FASTFORWARDSPACEFILE))
        gr_line = gr_line | np(PList('forward space file'), gr_yn, action=self._parse_setter(FORWARDSPACEFILE))
        gr_line = gr_line | np(PList('forward space record'), gr_yn, action=self._parse_setter(FORWARDSPACERECORD))
        gr_line = gr_line | np(PList('hardware end of medium'), gr_yn, action=self._parse_setter(HARDWAREENDOFMEDIUM))
        gr_line = gr_line | np(PList('label media'), gr_yn, action=self._parse_setter(LABELMEDIA))
        gr_line = gr_line | np(PList('offline on unmount'), gr_yn, action=self._parse_setter(OFFLINEONUNMOUNT))
        gr_line = gr_line | np(PList('random access'), gr_yn, action=self._parse_setter(RANDOMACCESS))
        gr_line = gr_line | np(PList('removable media'), gr_yn, action=self._parse_setter(REMOVABLEMEDIA))
        gr_line = gr_line | np(PList('two eof'), gr_yn, action=self._parse_setter(TWOEOF))
        gr_line = gr_line | np(PList('use mtiocget'), gr_yn, action=self._parse_setter(USEMTIOCGET))

        gr_res = OneOrMore(gr_line)

        result = gr_res.parseString(string, parseAll=True)
        if obj: self.link(obj)
        return 'Device: ' + self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Device suitable for inclusion in a
        configuration file.

        '''
        self.output = ['Device {\n  Name = "%(name)s"' % self,'}']
        
        for key in self.SETUP_KEYS: self._simple_phrase(key)
        for key in self.BOOL_KEYS: self._yesno_phrase(key)

        return '\n'.join(self.output)

# }}}
    # {{{ link(obj): link the device to a storage daemon

    def link(self, obj):
        '''Devices belong to Storage Daemons, but there's no intrinsic way to know
        that a device belongs to a daemon, so we have a table that provides
        links between the two.  This member, given a Storage object, will
        link the Device to the Storage.
        '''
        try:
            self.bc.do_sql(self._insert, (self[ID], obj[ID]))
        except Exception as e:
            if e.args[0] == 1062: pass # 1062 is what happens when you try to insert a duplicate row
            else:
                print(e)
                raise

            # }}}
    # {{{ unlink(obj): unlink the device from a storage daemon

    def unlink(self, obj):
        '''Remove the linkage between a Device and a Storage'''
        self.bc.do_sql(self._delete, (self[ID], obj[ID]))
        return

            # }}}
    # {{{ _cli_special_setup(): (un)link Storage

    def _cli_special_setup(self):
        '''Enable the CLI to (un)link devices to/from Sotrage Daemons.'''
        group = optparse.OptionGroup(self.parser, "Storage daemon links",
                                     "A device is associated with one or more storage daemons.")
        group.add_option('--add-link', metavar='STORAGE_DAEMON')
        group.add_option('--remove-link', metavar='STORAGE_DAEMON')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): (un)link Storage

    def _cli_special_do_parse(self, args):
        '''Handle any attempts by the CLI to (un)link the Device to/from Storage'''
        if args.add_link:
            s = bacula_tools.Storage().search(args.add_link)
            if not s[ID]: s.search(args.add_link)
            if not s[ID]:
                print('\n***WARNING***: Unable to find a Storage Daemon identified by "%s".  Not linked.\n' % args.add_link)
                return
            self.link(s)

        if args.remove_link:
            s = bacula_tools.Storage().search(args.remove_link)
            if not s[ID]: s.search(args.remove_link)
            if not s[ID]:
                print('\n***WARNING***: Unable to find a Storage Daemon identified by "%s".  Not unlinked.\n' % args.remove_link)
                return
            self.unlink(s)

        return

# }}}
    # {{{ _cli_special_print(): Display any linked Storage

    def _cli_special_print(self):
        '''Print any linked Storage'''
        for row in self.bc.do_sql(self._select, self[ID]):
            s = bacula_tools.Storage().search(row[0])
            print(('%'+ str(self._maxlen) + 's: %s') % ('Storage Daemon', s[NAME]))
        return

    # }}}
    # {{{ _cli_special_clone(oid): Handle cloning requirements.

    def _cli_special_clone(self, oid):
        '''Any clones of this device will be linked to the same Storage.'''
        for row in self.bc.do_sql(self._select, oid): self.bc.do_sql(self._insert, (self[ID], row[0]))
        return

        # }}}

def main():
    s = Device()
    s.cli()

if __name__ == "__main__": main()



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *
from re import compile, MULTILINE, IGNORECASE, DOTALL

class Schedule(DbDict):
    table = SCHEDULES
    # {{{ __init__(row={}, string=None):

    def __init__(self, row={}, string=None):
        '''Overrides DbDice.__init__ to ensure each instance has a clean entries member.'''
        DbDict.__init__(self, row, string)
        self.entries = []
        return

    # }}}
    # {{{ parse_string(string): update schedule from a string (config file entry)

    def parse_string(self, string):
        '''Populate a new object from a string.
        
        We're going to make a log of assumption here, not the least of
        which is that what will be passed in is a single Schedule{}
        resource, without comments.  Anything else will throw an
        exception.'''
        run_re = compile(r'^\s*run\s*=\s*(.*)', MULTILINE|IGNORECASE)
        data = string.strip()
        g = self.name_re.search(data).groups()
        s = g[0].strip()
        if s[0] == '"' and  s[-1] == '"': s = s[1:-1]
        if s[0] == "'" and  s[-1] == "'": s = s[1:-1]
        self._set_name(s)
        data = self.name_re.sub('', data)
        while True:
            g = run_re.search(data)
            if not g: break
            s = g.group(1).strip()
            if s[0] == '"' and  s[-1] == '"': s = s[1:-1]
            if s[0] == "'" and  s[-1] == "'": s = s[1:-1]
            self._add_run(s)
            data = run_re.sub('', data, 1)
        return "Schedule: " + self[NAME]

    # }}}
    # {{{ _add_run(run): add another run

    def _add_run(self, run):
        '''Add a schedule line for when jobs should run.'''
        new_run = self.bc.value_ensure(SCHEDULE_TIME, DATA, run)[0]
        self.entries.append(new_run)
        row = self.bc.do_sql('SELECT * FROM schedule_link WHERE scheduleid = %s AND timeid = %s', (self[ID], new_run[0]))
        if not row:
            self.bc.do_sql('INSERT INTO schedule_link(scheduleid, timeid) VALUES (%s, %s)', (self[ID], new_run[0]))

        return

    # }}}
    # {{{ _delete_run(run): add another run

    def _delete_run(self, run):
        '''Unlink a run line from this Schedule.  Does not remove unused lines from the database.'''
        for row in self.entries:
            if not row[1] == run: continue
            self.bc.do_sql('DELETE FROM schedule_link WHERE scheduleid = %s AND timeid = %s', (self[ID], row[0]))
            self.entries.remove(row)
            return
        print('I cannot delete Run entries that do not exist!')
        return

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Schedule suitable for inclustion in a config file.'''
        self.output = ['''Schedule {\n  Name = "%(name)s"''' % self, '}']
        [self.output.insert(-1,'  Run = %s' % x[1]) for x in self.entries]
        return '\n'.join(self.output)

# }}}
    # {{{ _load_runs(): helper for loading self from the database

    def _load_runs(self):
        '''When instantiating a Schedule, load all of the associated runs.'''
        if not self[ID]: return
        self.entries = list(self.bc.do_sql('''SELECT b.id AS id, b.data AS data FROM schedule_link a, schedule_time b
                                           WHERE a.scheduleid = %s AND a.timeid = b.id''', (self[ID],)))
        return self

    # }}}
    # {{{ _cli_special_setup(): setup the weird phrases that go with schedules

    def _cli_special_setup(self):
        '''Add CLI options for adding and removing run lines.'''
        group = optparse.OptionGroup(self.parser,
                                     "Stanzas",
                                     "Add or remove time windows from the schedule")
        group.add_option('--add', action='append', default=[], help='Add a time')
        group.add_option('--remove', action='append', default=[], help='Removes a time')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle the weird phrases that go with schedules

    def _cli_special_do_parse(self, args):
        '''Actually handle CLI options for adding and removing runs.'''
        for stanza in args.add: self._add_run(stanza)
        for stanza in args.remove: self._delete_run(stanza)
        return

# }}}
    # {{{ _cli_special_print(): print out the weird phrases that go with schedules

    def _cli_special_print(self):
        '''Print out the run lines for the CLI.'''
        print('Run times:')
        fmt = '	%s'
        [print(fmt % x[1]) for x in self.entries]
                
    # }}}
    # {{{ _cli_special_clone(oid):

    def _cli_special_clone(self, oid):
        '''Clones should start off with the same set of runs.'''
        select = 'SELECT %s,timeid FROM schedule_link WHERE scheduleid = %%s' % self[ID]
        insert = 'INSERT INTO schedule_link (scheduleid,timeid) VALUES (%s,%s)'
        for row in self.bc.do_sql(select, oid): self.bc.do_sql(insert, row)
        self._load_runs()
        return

# }}}

    
def main():
    s = Schedule()
    s.cli()

if __name__ == "__main__": main()



#! /usr/bin/env python

import ez_setup
ez_setup.use_setuptools()

import os, glob
from setuptools import setup, find_packages

NAME = 'bacula_configuration'
VERSION = '0.7'
WEBSITE = 'http://gallew.org/bacula_configuration'
LICENSE = 'GPLv3 or later'
DESCRIPTION = 'Bacula configuration management tool'
AUTHOR = 'Brian Gallew'
EMAIL = 'bacula_configuration@gallew.org'

setup(name = NAME,
      version = VERSION,
      description = DESCRIPTION,
      long_description = open('README.md').read(),
      author = AUTHOR,
      author_email = EMAIL,
      url = WEBSITE,
      install_requires = ['mysql-python'],
      extras_require = { 'parsing': ['pyparsing']},
      scripts = glob.glob('bin/*_configuration'),
      include_package_data = True,
      package_data = {
        'bacula_tools': ['data/*'],
        },
      packages = ['bacula_tools',],
      classifiers = ['Development Status :: 4 - Beta',
                     'Environment :: Console',
                     'Intended Audience :: System Administrators',
                     'Operating System :: OS Independent',
                     'Programming Language :: Python',
                     'License :: OSI Approved :: GNU General Public License (GPL)',
                     'Topic :: Utilities'],
      entry_points = {
          'console_scripts': [
              'manage_clients = bacula_tools.client:main',
              'manage_catalogs = bacula_tools.catalog:main',
              'manage_devices = bacula_tools.device:main',
              'manage_filesets = bacula_tools.fileset:main',
              'manage_pools = bacula_tools.pool:main',
              'manage_storage = bacula_tools.storage:main',
              'manage_messages = bacula_tools.messages:main',
              'manage_schedules = bacula_tools.schedule:main',
              'manage_jobs = bacula_tools.job:main',
              'manage_directors = bacula_tools.directors:main',
              'manage_scripts = bacula_tools.scripts:main',
              ]
      },
      )



#! /usr/bin/python
# Library to make writing Python tools for Bacula as easy as possible, with lots of code re-use.

# You must import both sys and bacula
import sys, bacula, subprocess
from random import choice
numarray = range(0,100)

# This is the list of Bacula daemon events that you
#  can receive.
class BaculaEvents(object):
    def __init__(self):
        # Called here when a new Bacula Events class is
        #  is created. Normally not used
        noop = 1
        return

    def JobStart(self, job):
        """
        Called here when a new job is started. If you want
        to do anything with the Job, you must register
        events you want to receive.
        """
        noop = 1
        return

  # Bacula Job is going to terminate
    def JobEnd(self, job):
        # If this is a verify job, bail
        if 'Verify' in job.Job: return
        if 'Snap' in job.Job: return # cannot verify snapshot backups

        # Let's truncate any new purged volumes in this pool.
        runcmd = 'purge volume action=Truncate allpools storage=%s\nquit\n' % job.Storage
        # We cannot use job.run() because it won't accept the action
        # keyword for purge.  Le sigh.
        try:
            so = subprocess.Popen('bconsole', shell=True,
                                  stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE).communicate(runcmd)[0]
        except Exception as e:
            job.JobReport = 'Bailing ... '
            job.JobReport = str(e)
            return
        job.JobReport = "ran volume purge command: %s" % so

        # This bit gives us a chance (5%) of running a verify job against this one.
        if choice(numarray) < 5:          # 5% chance of running a verify
            runcmd = 'run job=%s-Verify jobid=%d yes' % (job.Job, job.JobId)
            newjob = job.run(runcmd)
            job.JobReport="%s: jobid=%d\n" % (runcmd, newjob)
        return

  # Called here when the Bacula daemon is going to exit
    def Exit(self, job):
        noop = 1
        return

# This is what makes it actually work
bacula.set_events(BaculaEvents()) # register daemon events desired



#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

class Messages(DbDict):
    SETUP_KEYS = [(DATA, ''),]
    table = MESSAGES
    _select = 'SELECT ref_id, link_type FROM messages_link WHERE messages_id=%s'
    _insert = 'INSERT INTO messages_link (messages_id, ref_id, link_type) VALUES (%s, %s, %s)'
    _delete = 'DELETE FROM messages_link WHERE messages_id = %s AND ref_id=%s AND link_type=%s'
    # {{{ parse_string(string, obj=None):

    def parse_string(self, string, obj=None):
        '''Extend the standard parse_string functionality with object linkage.
        This is the solution I came up with for associating a Message with
        a Director, Client, or Storage object.
        '''
        retval = DbDict.parse_string(self, string)
        if obj: self.link(obj)
        return retval

        # }}}
    # {{{ link(obj):

    def link(self, obj):
        '''If I were willing to really wed this to MySQL, I should switch to REPLACE INTO.
        Since I'm not, we'll just do an insert and check for a duplication error.'''
        try:
            self.bc.do_sql(self._insert, (self[ID], obj[ID], obj.IDTAG))
        except Exception as e:
            if e.args[0] == 1062: pass # 1062 is what happens when you try to insert a duplicate row
            else:
                print(e)
                raise

                # }}}
    # {{{ unlink(obj): unlink the device from a storage daemon

    def unlink(self, obj):
        '''Remove the link between this Message and the given object.'''
        self.bc.do_sql(self._delete, (self[ID], obj[ID], obj.IDTAG))
        return

            # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation suitable for inclustion in any config file.'''
        output = 'Messages {\n  Name = "%(name)s"\n  %(data)s\n}' % self
        return output

# }}}
    # {{{ _cli_special_setup(): add in linkage support

    def _cli_special_setup(self):
        '''Add CLI switches to link objects to this.  The object-type is required
        as you could easily have the same name for Directors, Clients, and
        Storage.
        '''
        group = optparse.OptionGroup(self.parser, "Object links",
                                     "Messages are used by clients, storage daemons, and directors.")
        group.add_option('--add-link', metavar='STORAGE_DAEMON')
        group.add_option('--remove-link', metavar='STORAGE_DAEMON')
        group.add_option('--object-type', help='Optional device type to disambiguate the desired linkage')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle password parsing

    def _cli_special_do_parse(self, args):
        '''Handle the CLI switches for linkage.  Use the object type to provide a context
        for name lookup.'''
        obj = None
        if args.object_type: obj = getattr(bacula_tools, args.object_type.capitalize(), None)
            
        if args.add_link:
            if obj: target = obj().search(args.add_link)
            else:
                for key in [Client, Storage, Director]:
                    target = key().search(args.add_link)
                    if target[ID]: break
            if target[ID]: self.link(target)
            else: print('Unable to find anything named', args.add_link)

        if args.remove_link:
            if obj: target = obj().search(args.remove_link)
            else:
                for key in [Client, Storage, Director]:
                    target = key().search(args.remove_link)
                    if target[ID]: break
            if not target: print('Unable to find anything named', args.remove_link)
            else: self.unlink(target)

        return

# }}}
    # {{{ _cli_special_print(): print out passwords

    def _cli_special_print(self):
        '''Print out the linked objects.'''
        resultset = self.bc.do_sql(self._select, self[ID])
        if not resultset: return
        fmt = '%'+ str(self._maxlen) + 's'
        for row in resultset:
            for key in [Client, Storage, Director]:
                if key.IDTAG == row[1]: print(fmt % key().search(row[0])[NAME])
        return

    # }}}

def main():
    s = Messages()
    s.cli()

if __name__ == "__main__": main()




#! /usr/bin/env python

from __future__ import print_function
try: from . import *
except: from bacula_tools import *

# These are some shortcuts I put here just to make later code look a little cleaner.
RBJ = PList('Run Before Job')
RAJ = PList('Run After Job')
RAFJ = PList( 'Run After Failed Job')
CRBJ = PList('Client Run Before Job')
CRAJ = PList('Client Run After Job')

class Job(DbDict):
    '''This actually covers both Jobs and JobDefs, EXCEPT during parsing, when
    the JobDef subclass is used because it has different default values.
    '''
    SETUP_KEYS = [
        # Enum
        TYPE, LEVEL,
        # Strings
        NOTES, ADDPREFIX, ADDSUFFIX, BASE, BOOTSTRAP, MAXIMUMBANDWIDTH, MAXSTARTDELAY,
        REGEXWHERE, RUN, SPOOLSIZE, STRIPPREFIX, VERIFYJOB, WHERE, WRITEBOOTSTRAP,
        # keys with values
        (REPLACE, 'always'), 
        # Times
        DIFFERENTIALMAXWAITTIME, IDMAXWAITTIME, INCREMENTALMAXRUNTIME, MAXRUNSCHEDTIME,
        MAXRUNTIME, MAXWAITTIME, MAXFULLINTERVAL, RESCHEDULEINTERVAL, 
        ]
    INT_KEYS = [MAXIMUMCONCURRENTJOBS, RESCHEDULETIMES, PRIORITY]
    BOOL_KEYS = [ENABLED, PREFERMOUNTEDVOLUMES, ACCURATE, ALLOWDUPLICATEJOBS,
                 ALLOWMIXEDPRIORITY, CANCELLOWERLEVELDUPLICATES,
                 CANCELQUEUEDDUPLICATES, CANCELRUNNINGDUPLICATES, JOBDEF, 
                 PREFIXLINKS, PRUNEFILES, PRUNEJOBS, PRUNEVOLUMES, RERUNFAILEDLEVELS,
                 RESCHEDULEONERROR, SPOOLATTRIBUTES, SPOOLDATA, WRITEPARTAFTERJOB
             ]
    REFERENCE_KEYS = [          # foreign keys
        DIFFERENTIALPOOL_ID, FILESET_ID, FULLPOOL_ID, CLIENT_ID,
        INCREMENTALPOOL_ID, MESSAGES_ID, POOL_ID, SCHEDULE_ID, STORAGE_ID,
        ]        
    SPECIAL_KEYS = [JOB_ID,]    # These won't be handled en- masse
    table = JOBS
    retlabel = 'Job'
    # {{{ __init__(row={}, string = None):

    def __init__(self, row={}, string = None):
        '''Need to have a nice, clean scripts member'''
        DbDict.__init__(self, row, string)
        self.scripts = []
        return

    # }}}
    # {{{ parse_string(string): Entry point for a recursive descent parser

    def parse_string(self, string):
        # {{{ boilerplate.  Sigh

        '''Populate a new object from a string.
        
        Parsing is hard, so we're going to call out to the pyparsing
        library here.  I hope you installed it!
        FTR: this is hideous.
        '''
        from pyparsing import Suppress, Regex, quotedString, restOfLine, Keyword, nestedExpr, Group, OneOrMore, Word, Literal, alphanums, removeQuotes, replaceWith, nums, printables, nullDebugAction
        gr_eq = Literal('=')
        gr_stripped_string = quotedString.copy().setParseAction( removeQuotes )
        gr_opt_quoted_string = gr_stripped_string | restOfLine
        gr_number = Word(nums)
        gr_yn = Keyword('yes', caseless=True).setParseAction(replaceWith('1')) | Keyword('no', caseless=True).setParseAction(replaceWith('0'))

        def np(words, fn = gr_opt_quoted_string, action=nullDebugAction):
            p = Keyword(words[0], caseless=True).setDebug(bacula_tools.DEBUG)
            for w in words[1:]:
                p = p | Keyword(w, caseless=True).setDebug(bacula_tools.DEBUG)
            p = p + gr_eq + fn
            p.setParseAction(action)
            return p

        # }}}
        # {{{ # Easy ones that go don't take embedded spaces because I say so.

        gr_line = np((NAME,), action=lambda x: self._set_name(x[2]))
        for key in [TYPE, LEVEL, REPLACE, BASE, RUN, WHERE]:
            gr_line = gr_line | np((key,), action=self._parse_setter(key))

            # }}}
        # {{{ # Group of _id variables

        gr_line = gr_line | np(PList('differential pool'), action=self._parse_setter(DIFFERENTIALPOOL_ID))
        gr_line = gr_line | np(PList('file set'), action=self._parse_setter(FILESET_ID, dereference=True))
        gr_line = gr_line | np(PList('full pool'), action=self._parse_setter(FULLPOOL_ID))
        gr_line = gr_line | np((CLIENT,), action=self._parse_setter(CLIENT_ID, dereference=True))
        gr_line = gr_line | np(PList('incremental pool'), action=self._parse_setter(INCREMENTALPOOL_ID))
        gr_line = gr_line | np((MESSAGES,), action=self._parse_setter(MESSAGES_ID, dereference=True))
        gr_line = gr_line | np((POOL,), action=self._parse_setter(POOL_ID, dereference=True))
        gr_line = gr_line | np((SCHEDULE,), action=self._parse_setter(SCHEDULE_ID, dereference=True))
        gr_line = gr_line | np((STORAGE,), action=self._parse_setter(STORAGE_ID, dereference=True))
        gr_line = gr_line | np(PList('job defs'), action=self._parse_setter(JOB_ID, dereference=True))

        # }}}
        # {{{ # INTs

        gr_line = gr_line | np(PList('maximum concurrent jobs'), gr_number, action=self._parse_setter(MAXIMUMCONCURRENTJOBS))
        gr_line = gr_line | np(PList('re schedule times'), gr_number, action=self._parse_setter(RESCHEDULETIMES))
        gr_line = gr_line | np((PRIORITY,), gr_number, action=self._parse_setter(PRIORITY))

        # }}}
        # {{{ # True/False

        gr_line = gr_line | np((ACCURATE,), gr_yn, action=self._parse_setter(ACCURATE))
        gr_line = gr_line | np(PList('allow duplicate jobs'), gr_yn, action=self._parse_setter(ALLOWDUPLICATEJOBS))
        gr_line = gr_line | np(PList('allow mixed priority'), gr_yn, action=self._parse_setter(ALLOWMIXEDPRIORITY))
        gr_line = gr_line | np(PList('cancel lower level duplicates'), gr_yn, action=self._parse_setter(CANCELLOWERLEVELDUPLICATES))
        gr_line = gr_line | np(PList('cancel queued duplicates'), gr_yn, action=self._parse_setter(CANCELQUEUEDDUPLICATES))
        gr_line = gr_line | np(PList('cancel running duplicates'), gr_yn, action=self._parse_setter(CANCELRUNNINGDUPLICATES))
        gr_line = gr_line | np((ENABLED,), gr_yn, action=self._parse_setter(ENABLED))
        gr_line = gr_line | np(PList('prefer mounted volumes'), gr_yn, action=self._parse_setter(PREFERMOUNTEDVOLUMES))
        gr_line = gr_line | np(PList('prefix links'), gr_yn, action=self._parse_setter(PREFIXLINKS))
        gr_line = gr_line | np(PList('prune files'), gr_yn, action=self._parse_setter(PRUNEFILES))
        gr_line = gr_line | np(PList('prune jobs'), gr_yn, action=self._parse_setter(PRUNEJOBS))
        gr_line = gr_line | np(PList('prune volumes'), gr_yn, action=self._parse_setter(PRUNEVOLUMES))
        gr_line = gr_line | np(PList('re run failed levels'), gr_yn, action=self._parse_setter(RERUNFAILEDLEVELS))
        gr_line = gr_line | np(PList('re schedule on error'), gr_yn, action=self._parse_setter(RESCHEDULEONERROR))
        gr_line = gr_line | np(PList('spool attributes'), gr_yn, action=self._parse_setter(SPOOLATTRIBUTES))
        gr_line = gr_line | np(PList('spool data'), gr_yn, action=self._parse_setter(SPOOLDATA))
        gr_line = gr_line | np(PList('write boot strap'), gr_yn, action=self._parse_setter(WRITEPARTAFTERJOB))

        # }}}
        # {{{ # plain strings

        gr_line = gr_line | np((NOTES,), action=self._parse_setter(NOTES))
        gr_line = gr_line | np((ADDPREFIX, 'add prefix'), action=self._parse_setter(ADDPREFIX))
        gr_line = gr_line | np((ADDSUFFIX, 'add suffix'), action=self._parse_setter(ADDSUFFIX))
        gr_line = gr_line | np((BASE,), action=self._parse_setter(BASE))
        gr_line = gr_line | np((BOOTSTRAP, 'boot strap'), action=self._parse_setter(BOOTSTRAP))
        gr_line = gr_line | np((DIFFERENTIALMAXWAITTIME, 'differential max wait time', 'differentialmaxwait time', 'differentialmax waittime', 'differential maxwaittime', 'differentialmax wait time', 'differential maxwait time', 'differential max waittime'), action=self._parse_setter(DIFFERENTIALMAXWAITTIME))
        gr_line = gr_line | np(('incremental-differentialmaxwaittime','incremental-differential maxwaittime','incremental-differentialmax waittime','incremental-differentialmaxwait time','incremental-differential max waittime','incremental-differential maxwait time','incremental-differentialmax wait time','incremental-differential max wait time',), action=self._parse_setter(IDMAXWAITTIME))
        gr_line = gr_line | np((INCREMENTALMAXRUNTIME, 'incremental max run time', 'incrementalmaxrun time', 'incrementalmax runtime', 'incremental maxruntime', 'incrementalmax run time', 'incremental maxrun time', 'incremental max runtime'), action=self._parse_setter(INCREMENTALMAXRUNTIME))
        gr_line = gr_line | np((MAXFULLINTERVAL, 'max full interval', 'max fullinterval', 'maxfull interval'), action=self._parse_setter(MAXFULLINTERVAL))
        gr_line = gr_line | np((MAXIMUMBANDWIDTH, 'maximum band width', 'maximum bandwidth', 'maximumband width'), action=self._parse_setter(MAXIMUMBANDWIDTH))
        gr_line = gr_line | np((MAXRUNSCHEDTIME, 'max run sched time', 'maxrunsched time', 'maxrun schedtime', 'max runschedtime', 'maxrun sched time', 'max runsched time', 'max run schedtime'), action=self._parse_setter(MAXRUNSCHEDTIME))
        gr_line = gr_line | np((MAXRUNTIME, 'max run time', 'maxrun time', 'max runtime'), action=self._parse_setter(MAXRUNTIME))
        gr_line = gr_line | np((MAXSTARTDELAY, 'max start delay', 'max startdelay', 'maxstart delay'), action=self._parse_setter(MAXSTARTDELAY))
        gr_line = gr_line | np((MAXWAITTIME, 'max wait time', 'max waittime', 'maxwait time'), action=self._parse_setter(MAXWAITTIME))
        gr_line = gr_line | np((REGEXWHERE, 'regex where'), action=self._parse_setter(REGEXWHERE))
        gr_line = gr_line | np((RESCHEDULEINTERVAL, 're schedule interval', 're scheduleinterval', 'reschedule interval'), action=self._parse_setter(RESCHEDULEINTERVAL))
        gr_line = gr_line | np((RUN,), action=self._parse_setter(RUN))
        gr_line = gr_line | np((SPOOLSIZE, 'spool size'), action=self._parse_setter(SPOOLSIZE))
        gr_line = gr_line | np((STRIPPREFIX, 'strip prefix'), action=self._parse_setter(STRIPPREFIX))
        gr_line = gr_line | np((VERIFYJOB, 'verify job'), action=self._parse_setter(VERIFYJOB))
        gr_line = gr_line | np((WHERE,), action=self._parse_setter(WHERE))
        gr_line = gr_line | np((WRITEBOOTSTRAP, 'write boot strap', 'write bootstrap', 'writeboot strap'), action=self._parse_setter(WRITEBOOTSTRAP))

        # }}}
        # The ugliness that is run scripts
        gr_line = gr_line | np(RBJ, gr_stripped_string,
                               action=self._parse_script(runsonclient=0, runswhen='Before'))
        gr_line = gr_line | np(RAJ, gr_stripped_string,
                               action=self._parse_script(runsonclient=0, runswhen='After'))
        gr_line = gr_line | np(RAFJ, gr_stripped_string,
                               action=self._parse_script(runsonsuccess=0, runsonfailure=1,
                                                         runsonclient=0, runswhen='After'))
        gr_line = gr_line | np(CRBJ, gr_stripped_string,
                               action=self._parse_script(runswhen='Before'))
        gr_line = gr_line | np(CRAJ, gr_stripped_string,
                               action=self._parse_script(runswhen='After'))
        
        # This is a complicated one
        gr_script_parts = np(('Command',), gr_stripped_string)
        gr_script_parts = gr_script_parts | np((CONSOLE,), gr_stripped_string)
        gr_script_parts = gr_script_parts | np(PList('Runs When'))
        gr_script_parts = gr_script_parts | np(PList('Runs On Success'), gr_yn)
        gr_script_parts = gr_script_parts | np(PList('Runs On Failure'), gr_yn)
        gr_script_parts = gr_script_parts | np(PList('Runs On Client'), gr_yn)
        gr_script_parts = gr_script_parts | np(PList('Fail Job On Error'), gr_yn)
        gr_script = ((Keyword('Run Script', caseless=True) | Keyword('RunScript', caseless=True)) + nestedExpr('{','}', OneOrMore(gr_script_parts))).setParseAction(self._parse_script_full)
        
        gr_res = OneOrMore(gr_line | gr_script)
        try:
            result = gr_res.parseString(string, parseAll=True)
        except Exception as e:
            print(e)
            raise
        return self.retlabel + ': '+ self[NAME]

    # }}}
    # {{{ __str__(): 

    def __str__(self):
        '''String representation of a Job, suitable for inclusion in a Director config'''
        self.output = ['%s {' % self.retlabel,'}']
        for x in self.SETUP_KEYS: self._simple_phrase(x)
        for x in self.INT_KEYS: self._simple_phrase(x)
        for x in self.REFERENCE_KEYS:
            if self[x] == None: continue
            self.output.insert(-1,'  %s = "%s"' % (x.replace('_id', '').capitalize(), self._fk_reference(x)[NAME]))
        if self[JOB_ID]: self.output.insert(-1,'  JobDefs = "%s"' % self._fk_reference(JOB_ID)[NAME])

        for x in self.BOOL_KEYS:
            if x == JOBDEF: continue
            self._yesno_phrase(x)
        for x in self.scripts: self.output.insert(-1, str(x))
        return '\n'.join(self.output)

# }}}
    # {{{ _fk_reference(fk, string=None): Set/get fk-references

    def _fk_reference(self, fk, string=None):
        '''This overrides the normal _fk_reference function becase we actually have
        four different keys that all point to Pools.
        '''
        key = fk.replace('_id', '')
        if 'pool' in key: key = 'pool'
        obj = bacula_tools._DISPATCHER[key]()
        if string:
            obj.search(string.strip())
            if not obj[ID]: obj._set_name(string.strip())
            if not self[fk] == obj[ID]: self._set(fk, obj[ID])
        else: obj.search(self[fk])
        return obj

# }}}
    # {{{ _load_scripts():

    def _load_scripts(self):
        '''Job scripts are stored separately as Script objects.  This loads them in. '''
        if self[ID]:
            for row in self.bc.do_sql('SELECT * FROM job_scripts WHERE job_id = %s', (self[ID],), asdict=True):
                s = bacula_tools.Script({ID: row[SCRIPT_ID]})
                s.search()
                self.scripts = [x for x in self.scripts if not x[ID] == s[ID]]
                self.scripts.append(s)
        return

    # }}}
    # {{{ _parse_script(**kwargs): returns a parser for the script shortcuts

    def _parse_script(self, **kwargs):
        '''Helper function for parsing configuration strings.'''
        def doit(a,b,c):
            s = bacula_tools.Script(kwargs)
            s[COMMAND] = c[2]
            s.search()
            return self._add_script(s)
        return doit

    # }}}
    # {{{ _add_script(s): Add a script to myself

    def _add_script(self, s):
        '''Add a script to the Job.'''
        self.scripts = [x for x in self.scripts if not x[ID] == s[ID]]
        self.scripts.append(s)
        row = self.bc.do_sql('SELECT * FROM job_scripts WHERE job_id = %s AND script_id = %s', (self[ID], s[ID]))
        if not row:
            self.bc.do_sql('INSERT INTO job_scripts(job_id, script_id) VALUES (%s, %s)', (self[ID], s[ID]))
        return s

    # }}}
    # {{{ _delete_script(s): Remove a script from myself

    def _delete_script(self, s):
        '''Remove a Script from the Job.  This does not actually delete the Script,
        just the linkage to this job.'''
        self.bc.do_sql('DELETE FROM job_scripts WHERE id = %s', (s[ID]))
        self.scripts = [x for x in self.scripts if not x[ID] == s[ID]]
        return

    # }}}
    # {{{ _parse_script_full(tokens):

    def _parse_script_full(self, *tokens):
        '''Another helper for script parsing.'''
        from pprint import pprint
        s = bacula_tools.Script()
        values = tokens[2][1]
        while values:
            k,n,v = values[:3]
            del values[:3]
            s[k.lower()] = v
        self._add_script(s.search())
        return

# }}}
    # {{{ _cli_special_setup(): setup the weird phrases that go with jobs

    def _cli_special_setup(self):
        '''Suport for adding all of the foreign-key references to the CLI.'''
        group = optparse.OptionGroup(self.parser,
                                     "Object Setters",
                                     "Various objects associated with a Job")
        group.add_option('--pool', help='Use this pool for all backups unless overridden by a more specific pool')
        group.add_option('--differential-pool', help='Use this pool for differential backups instead of the standard pool')
        group.add_option('--full-pool', help='Use this pool for full backups instead of the standard pool')
        group.add_option('--incremental-pool', help='Use this pool for incremental backups instead of the standard pool')
        group.add_option('--fileset')
        group.add_option('--client')
        group.add_option('--message-set')
        group.add_option('--schedule')
        group.add_option('--storage')
        group.add_option('--default-job', help='The job which will supply default values for those otherwise unset on this one')
        self.parser.add_option_group(group)
        return

    # }}}
    # {{{ _cli_special_do_parse(args): handle the weird phrases that go with jobs

    def _cli_special_do_parse(self, args):
        '''CLI Foreign Key reference actions.'''
        self._cli_deref_helper(POOL_ID, args.pool, Pool)
        self._cli_deref_helper(DIFFERENTIALPOOL_ID, args.differential_pool, Pool)
        self._cli_deref_helper(FULLPOOL_ID, args.full_pool, Pool)
        self._cli_deref_helper(INCREMENTALPOOL_ID, args.incremental_pool, Pool)

        self._cli_deref_helper(FILESET_ID, args.fileset, Fileset)
        self._cli_deref_helper(CLIENT_ID, args.client, Client)
        self._cli_deref_helper(MESSAGES_ID, args.message_set, Messages)
        self._cli_deref_helper(SCHEDULE_ID, args.schedule, Schedule)
        self._cli_deref_helper(STORAGE_ID, args.storage, Storage)
        self._cli_deref_helper(JOB_ID, args.default_job, JobDef)
        return

    def _cli_deref_helper(self, key, value, obj):
        '''Shortcut function to make _cli_special_do_parse() a lot cleaner.'''
        if value == None: return
        if value=='': return self._set(key, None)
        target = obj().search(value)
        if target[ID]: self._set(key, target[ID])
        else: print('Unable to find a match for %s, continuing' % value)
        pass
# }}}
    # {{{ _cli_special_print(): print out the weird phrases that go with jobs

    def _cli_special_print(self):
        '''All of the foreign key objects get printed out here for the CLI.'''
        fmt = '%' + str(self._maxlen) + 's: %s'
        for x in self.REFERENCE_KEYS:
            if self[x] == None: continue
            print(fmt % (x.replace('_id', '').capitalize(), self._fk_reference(x)[NAME]))
        if self.scripts:
            print( '\nScripts')
            for x in self.scripts: print( x)
        return
    # }}}
    # {{{ _cli_special_clone(oid): script links for CLI

    def _cli_special_clone(self, oid):
        '''When cloning, add in script links.'''
        select = 'SELECT %s,script_id FROM job_scripts WHERE job_id = %%s' % self[ID]
        insert = 'INSERT INTO job_scripts (job_id,script_id) VALUES (%s,%s)'
        for row in self.bc.do_sql(select, oid): self.bc.do_sql(insert, row)
        self._load_scripts()
        pass

        # }}}

class JobDef(Job):
    '''This is really just a Job with a different label (for printing) and a value of 1 for the JOBDEF key.'''
    retlabel = 'JobDefs'
    
    def _save(self):
        '''JobDefs force the JOBDEF key to 1 upon saving.'''
        self[JOBDEF] = 1
        return Job._save(self)

def main():
    s = Job()
    s.cli()

if __name__ == "__main__": main()
